# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import os.path as osp
import tempfile
from os import remove

import mmcv
import numpy as np
import pytest

from otx.algorithms.segmentation.adapters.mmseg.utils.data_utils import (
    abs_path_if_valid,
    add_labels,
    check_labels,
    create_annotation_from_hard_seg_map,
    get_classes_from_annotation,
    get_extended_label_names,
    load_dataset_items,
    load_labels_from_annotation,
)
from otx.api.entities.label import Domain, LabelEntity
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


def _create_dummy_coco_json(json_name):
    image = {
        "id": 0,
        "width": 640,
        "height": 640,
        "file_name": "fake_name.jpg",
    }

    annotation_1 = {
        "id": 1,
        "image_id": 0,
        "category_id": 0,
        "area": 400,
        "bbox": [50, 60, 20, 20],
        "iscrowd": 0,
    }

    annotation_2 = {
        "id": 2,
        "image_id": 0,
        "category_id": 0,
        "area": 900,
        "bbox": [100, 120, 30, 30],
        "iscrowd": 0,
    }

    categories = [
        {
            "id": 0,
            "name": "car",
            "supercategory": "car",
        }
    ]

    fake_json = {
        "images": [image],
        "annotations": [annotation_1, annotation_2],
        "categories": categories,
    }

    mmcv.dump(fake_json, json_name)


class TestMMSegDataUtilsInputParamsValidation:
    @e2e_pytest_unit
    def test_abs_path_if_valid_input_params_validation(self):
        """
        <b>Description:</b>
        Check "abs_path_if_valid" function input parameters validation

        <b>Input data:</b>
        "value" unexpected object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "abs_path_if_valid" function
        """
        for unexpected_value in [
            # Unexpected integer is specified as "value" parameter
            1,
            # Empty string is specified as "value" parameter
            "",
            # Path with null character is specified as "value" parameter
            "./\0null_char",
            # Path with non-printable character is specified as "value" parameter
            "./\non_printable_char",
        ]:
            with pytest.raises(ValueError):
                abs_path_if_valid(value=unexpected_value)

    @e2e_pytest_unit
    def test_get_classes_from_annotation_input_params_validation(self):
        """
        <b>Description:</b>
        Check "get_classes_from_annotation" function input parameters validation

        <b>Input data:</b>
        "annot_path" unexpected object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "get_classes_from_annotation" function
        """
        for unexpected_value in [
            # non string object is specified as "annot_path" parameter
            1,
            # Empty string is specified as "annot_path" parameter
            "",
            # Path to file with unexpected extension is specified as "annot_path" parameter
            "./unexpected_extension.yaml",
            # Path to non-existing file is specified as "annot_path" parameter
            "./non_existing.json",
            # Path with null character is specified as "annot_path" parameter
            "./null\0char.json",
            # Path with non-printable character is specified as "annot_path" parameter
            "./\non_printable_char.json",
        ]:
            with pytest.raises(ValueError):
                get_classes_from_annotation(annot_path=unexpected_value)

    @e2e_pytest_unit
    def test_create_annotation_from_hard_seg_map_input_params_validation(self):
        """
        <b>Description:</b>
        Check "create_annotation_from_hard_seg_map" function input parameters validation

        <b>Input data:</b>
        "create_annotation_from_hard_seg_map" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "create_annotation_from_hard_seg_map" function
        """
        label = LabelEntity(name="test label", domain=Domain.SEGMENTATION)
        correct_values_dict = {
            "hard_seg_map": np.random.rand(2, 2),
            "labels": [label],
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "dataset_item" parameter
            ("hard_seg_map", unexpected_int),
            # Unexpected integer is specified as "labels" parameter
            ("labels", unexpected_int),
            # Unexpected integer is specified as nested label
            ("labels", [label, unexpected_int]),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=create_annotation_from_hard_seg_map,
        )

    @e2e_pytest_unit
    def test_load_labels_from_annotation_input_params_validation(self):
        """
        <b>Description:</b>
        Check "load_labels_from_annotation" function input parameters validation

        <b>Input data:</b>
        "ann_dir" unexpected object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "load_labels_from_annotation" function
        """
        for unexpected_value in [
            # Unexpected integer is specified as "ann_dir" parameter
            1,
            # Empty string is specified as "ann_dir" parameter
            "",
            # Path with null character is specified as "ann_dir" parameter
            "./\0null_char",
            # Path with non-printable character is specified as "ann_dir" parameter
            "./\non_printable_char",
        ]:
            with pytest.raises(ValueError):
                load_labels_from_annotation(ann_dir=unexpected_value)

    @e2e_pytest_unit
    def test_add_labels_input_params_validation(self):
        """
        <b>Description:</b>
        Check "add_labels" function input parameters validation

        <b>Input data:</b>
        "add_labels" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "add_labels" function
        """
        label = LabelEntity(name="test label", domain=Domain.SEGMENTATION)
        correct_values_dict = {
            "cur_labels": [label],
            "new_labels": [("label_name1", "label_id1")],
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "cur_labels" parameter
            ("cur_labels", unexpected_int),
            # Unexpected integer is specified as nested label
            ("cur_labels", [label, unexpected_int]),
            # Unexpected integer is specified as "new_labels" parameter
            ("new_labels", unexpected_int),
            # Unexpected integer is specified as nested new_label
            ("new_labels", [("label_name1", "label_id1"), unexpected_int]),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=add_labels,
        )

    @e2e_pytest_unit
    def test_check_labels_input_params_validation(self):
        """
        <b>Description:</b>
        Check "check_labels" function input parameters validation

        <b>Input data:</b>
        "check_labels" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "check_labels" function
        """
        label = LabelEntity(name="test label", domain=Domain.SEGMENTATION)
        correct_values_dict = {
            "cur_labels": [label],
            "new_labels": [("label_name1", "label_id1")],
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "cur_labels" parameter
            ("cur_labels", unexpected_int),
            # Unexpected integer is specified as nested label
            ("cur_labels", [label, unexpected_int]),
            # Unexpected integer is specified as "new_labels" parameter
            ("new_labels", unexpected_int),
            # Unexpected integer is specified as nested new_label
            ("new_labels", [("label_name1", "label_id1"), unexpected_int]),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=check_labels,
        )

    @e2e_pytest_unit
    def test_get_extended_label_names_input_params_validation(self):
        """
        <b>Description:</b>
        Check "get_extended_label_names" function input parameters validation

        <b>Input data:</b>
        "labels" unexpected object

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as input parameter for
        "get_extended_label_names" function
        """
        label = LabelEntity(name="test label", domain=Domain.SEGMENTATION)
        unexpected_int = 1
        for unexpected_value in [
            # Unexpected integer is specified as "labels" parameter
            unexpected_int,
            # Empty string is specified as nested label
            [label, unexpected_int],
        ]:
            with pytest.raises(ValueError):
                get_extended_label_names(labels=unexpected_value)

    @e2e_pytest_unit
    def test_load_dataset_items_params_validation(self):
        """
        <b>Description:</b>
        Check "load_dataset_items" function input parameters validation

        <b>Input data:</b>
        "load_dataset_items" function unexpected-type input parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "load_dataset_items" function
        """
        tmp_dir = tempfile.TemporaryDirectory()
        fake_json_file = osp.join(tmp_dir.name, "fake_data.json")
        _create_dummy_coco_json(fake_json_file)
        label = LabelEntity(name="test label", domain=Domain.DETECTION)
        correct_values_dict = {
            "ann_file_path": fake_json_file,
            "data_root_dir": tmp_dir.name,
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "ann_file_path" parameter
            ("ann_file_path", unexpected_int),
            # Empty string is specified as "ann_file_path" parameter
            ("ann_file_path", ""),
            # Path with null character is specified as "ann_file_path" parameter
            ("ann_file_path", osp.join(tmp_dir.name, "\0fake_data.json")),
            # Path with non-printable character is specified as "ann_file_path" parameter
            ("ann_file_path", osp.join(tmp_dir.name, "\nfake_data.json")),
            # Unexpected integer is specified as "data_root_dir" parameter
            ("data_root_dir", unexpected_int),
            # Empty string is specified as "data_root_dir" parameter
            ("data_root_dir", ""),
            # Path with null character is specified as "data_root_dir" parameter
            ("data_root_dir", "./\0null_char"),
            # Path with non-printable character is specified as "data_root_dir" parameter
            ("data_root_dir", "./\non_printable_char"),
            # Unexpected integer is specified as "subset" parameter
            ("subset", unexpected_int),
            # Unexpected integer is specified as "labels_list" parameter
            ("labels_list", unexpected_int),
            # Unexpected integer is specified as nested label
            ("labels_list", [label, unexpected_int]),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=load_dataset_items,
        )
        remove(fake_json_file)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.runner import force_fp32


class MixLossMixin(nn.Module):
    @staticmethod
    def _mix_loss(logits, target, ignore_index=255):
        num_samples = logits.size(0)
        assert num_samples % 2 == 0

        with torch.no_grad():
            probs = F.softmax(logits, dim=1)
            probs_a, probs_b = torch.split(probs, num_samples // 2)
            mean_probs = 0.5 * (probs_a + probs_b)
            trg_probs = torch.cat([mean_probs, mean_probs], dim=0)

        log_probs = torch.log_softmax(logits, dim=1)
        losses = torch.sum(trg_probs * log_probs, dim=1).neg()

        valid_mask = target != ignore_index
        valid_losses = torch.where(valid_mask, losses, torch.zeros_like(losses))

        return valid_losses.mean()

    @force_fp32(apply_to=("seg_logit",))
    def losses(self, seg_logit, seg_label, train_cfg, *args, **kwargs):
        loss = super().losses(seg_logit, seg_label, train_cfg, *args, **kwargs)
        if train_cfg.get("mix_loss", None) and train_cfg.mix_loss.get("enable", False):
            mix_loss = self._mix_loss(seg_logit, seg_label, ignore_index=self.ignore_index)

            mix_loss_weight = train_cfg.mix_loss.get("weight", 1.0)
            loss["loss_mix"] = mix_loss_weight * mix_loss

        return loss



# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import List, Sequence, Tuple, Union

import torch
import torch.nn.functional as F
from mmcls.models.necks.gap import GlobalAveragePooling

from otx.mpa.modules.models.heads.custom_atss_head import CustomATSSHead
from otx.mpa.modules.models.heads.custom_ssd_head import CustomSSDHead
from otx.mpa.modules.models.heads.custom_vfnet_head import CustomVFNetHead
from otx.mpa.modules.models.heads.custom_yolox_head import CustomYOLOXHead


class BaseRecordingForwardHook(ABC):
    """While registered with the designated PyTorch module, this class caches feature vector during forward pass.
    Example::
        with BaseRecordingForwardHook(model.module.backbone) as hook:
            with torch.no_grad():
                result = model(return_loss=False, **data)
            print(hook.records)
    Args:
        module (torch.nn.Module): The PyTorch module to be registered in forward pass
        fpn_idx (int, optional): The layer index to be processed if the model is a FPN.
                                  Defaults to 0 which uses the largest feature map from FPN.
    """

    def __init__(self, module: torch.nn.Module, fpn_idx: int = -1) -> None:
        self._module = module
        self._handle = None
        self._records = []
        self._fpn_idx = fpn_idx

    @property
    def records(self):
        return self._records

    @abstractmethod
    def func(x: torch.Tensor, fpn_idx: int = -1) -> torch.Tensor:
        """This method get the feature vector or saliency map from the output of the module.
        Args:
            x (torch.Tensor): Feature map from the backbone module
            fpn_idx (int, optional): The layer index to be processed if the model is a FPN.
                                    Defaults to 0 which uses the largest feature map from FPN.
        Returns:
            torch.Tensor (torch.Tensor): Saliency map for feature vector
        """
        raise NotImplementedError

    def _recording_forward(self, _: torch.nn.Module, input: torch.Tensor, output: torch.Tensor):
        tensors = self.func(output)
        tensors = tensors.detach().cpu().numpy()
        for tensor in tensors:
            self._records.append(tensor)

    def __enter__(self) -> BaseRecordingForwardHook:
        self._handle = self._module.backbone.register_forward_hook(self._recording_forward)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._handle.remove()


class EigenCamHook(BaseRecordingForwardHook):
    @staticmethod
    def func(feature_map: Union[torch.Tensor, Sequence[torch.Tensor]], fpn_idx: int = -1) -> torch.Tensor:
        if isinstance(feature_map, (list, tuple)):
            feature_map = feature_map[fpn_idx]

        x = feature_map.type(torch.float)
        bs, c, h, w = x.size()
        reshaped_fmap = x.reshape((bs, c, h * w)).transpose(1, 2)
        reshaped_fmap = reshaped_fmap - reshaped_fmap.mean(1)[:, None, :]
        U, S, V = torch.linalg.svd(reshaped_fmap, full_matrices=True)
        saliency_map = (reshaped_fmap @ V[:, 0][:, :, None]).squeeze(-1)
        max_values, _ = torch.max(saliency_map, -1)
        min_values, _ = torch.min(saliency_map, -1)
        saliency_map = 255 * (saliency_map - min_values[:, None]) / ((max_values - min_values + 1e-12)[:, None])
        saliency_map = saliency_map.reshape((bs, h, w))
        saliency_map = saliency_map.to(torch.uint8)
        return saliency_map


class ActivationMapHook(BaseRecordingForwardHook):
    @staticmethod
    def func(feature_map: Union[torch.Tensor, Sequence[torch.Tensor]], fpn_idx: int = -1) -> torch.Tensor:
        """Generate the saliency map by average feature maps then normalizing to (0, 255)."""
        if isinstance(feature_map, (list, tuple)):
            assert fpn_idx < len(
                feature_map
            ), f"fpn_idx: {fpn_idx} is out of scope of feature_map length {len(feature_map)}!"
            feature_map = feature_map[fpn_idx]

        bs, c, h, w = feature_map.size()
        activation_map = torch.mean(feature_map, dim=1)
        activation_map = activation_map.reshape((bs, h * w))
        max_values, _ = torch.max(activation_map, -1)
        min_values, _ = torch.min(activation_map, -1)
        activation_map = 255 * (activation_map - min_values[:, None]) / (max_values - min_values + 1e-12)[:, None]
        activation_map = activation_map.reshape((bs, h, w))
        activation_map = activation_map.to(torch.uint8)
        return activation_map


class FeatureVectorHook(BaseRecordingForwardHook):
    @staticmethod
    def func(feature_map: Union[torch.Tensor, Sequence[torch.Tensor]]) -> torch.Tensor:
        """Generate the feature vector by average pooling feature maps."""
        if isinstance(feature_map, (list, tuple)):
            # aggregate feature maps from Feature Pyramid Network
            feature_vector = [torch.nn.functional.adaptive_avg_pool2d(f, (1, 1)) for f in feature_map]
            feature_vector = torch.cat(feature_vector, 1)
        else:
            feature_vector = torch.nn.functional.adaptive_avg_pool2d(feature_map, (1, 1))
        return feature_vector


class DetSaliencyMapHook(BaseRecordingForwardHook):
    """Saliency map hook for object detection models."""

    def __init__(self, module: torch.nn.Module) -> None:
        super().__init__(module)
        self._neck = module.neck if module.with_neck else None
        self._bbox_head = module.bbox_head
        self._num_cls_out_channels = module.bbox_head.cls_out_channels  # SSD-like heads also have background class
        if hasattr(module.bbox_head, "anchor_generator"):
            self._num_anchors = module.bbox_head.anchor_generator.num_base_anchors
        else:
            self._num_anchors = [1] * 10

    def func(
        self,
        x: Union[torch.Tensor, List[torch.Tensor], Tuple[torch.Tensor]],
        _: int = -1,
        cls_scores_provided: bool = False,
    ) -> torch.Tensor:
        """
        Generate the saliency map from raw classification head output, then normalizing to (0, 255).

        :param x: Feature maps from backbone/FPN or classification scores from cls_head
        :param cls_scores_provided: If True - use 'x' as is, otherwise forward 'x' through the classification head
        :return: Class-wise Saliency Maps. One saliency map per each class - [batch, class_id, H, W]
        """
        if cls_scores_provided:
            cls_scores = x
        else:
            cls_scores = self._get_cls_scores_from_feature_map(x)

        bs, _, h, w = cls_scores[-1].size()
        saliency_maps = torch.empty(bs, self._num_cls_out_channels, h, w)
        for batch_idx in range(bs):
            cls_scores_anchorless = []
            for scale_idx, cls_scores_per_scale in enumerate(cls_scores):
                cls_scores_anchor_grouped = cls_scores_per_scale[batch_idx].reshape(
                    self._num_anchors[scale_idx], (self._num_cls_out_channels), *cls_scores_per_scale.shape[-2:]
                )
                cls_scores_out, _ = cls_scores_anchor_grouped.max(dim=0)
                cls_scores_anchorless.append(cls_scores_out.unsqueeze(0))
            cls_scores_anchorless_resized = []
            for cls_scores_anchorless_per_level in cls_scores_anchorless:
                cls_scores_anchorless_resized.append(
                    F.interpolate(cls_scores_anchorless_per_level, (h, w), mode="bilinear")
                )
            saliency_maps[batch_idx] = torch.cat(cls_scores_anchorless_resized, dim=0).mean(dim=0)

        saliency_maps = saliency_maps.reshape((bs, self._num_cls_out_channels, -1))
        max_values, _ = torch.max(saliency_maps, -1)
        min_values, _ = torch.min(saliency_maps, -1)
        saliency_maps = 255 * (saliency_maps - min_values[:, :, None]) / (max_values - min_values + 1e-12)[:, :, None]
        saliency_maps = saliency_maps.reshape((bs, self._num_cls_out_channels, h, w))
        saliency_maps = saliency_maps.to(torch.uint8)
        return saliency_maps

    def _get_cls_scores_from_feature_map(self, x: torch.Tensor) -> List:
        """Forward features through the classification head of the detector."""
        with torch.no_grad():
            if self._neck is not None:
                x = self._neck(x)

            if isinstance(self._bbox_head, CustomSSDHead):
                cls_scores = []
                for feat, cls_conv in zip(x, self._bbox_head.cls_convs):
                    cls_scores.append(cls_conv(feat))
            elif isinstance(self._bbox_head, CustomATSSHead):
                cls_scores = []
                for cls_feat in x:
                    for cls_conv in self._bbox_head.cls_convs:
                        cls_feat = cls_conv(cls_feat)
                    cls_score = self._bbox_head.atss_cls(cls_feat)
                    cls_scores.append(cls_score)
            elif isinstance(self._bbox_head, CustomVFNetHead):
                # Not clear how to separate cls_scores from bbox_preds
                cls_scores, _, _ = self._bbox_head(x)
            elif isinstance(self._bbox_head, CustomYOLOXHead):

                def forward_single(x, cls_convs, conv_cls):
                    """Forward feature of a single scale level."""
                    cls_feat = cls_convs(x)
                    cls_score = conv_cls(cls_feat)
                    return cls_score

                map_results = map(
                    forward_single, x, self._bbox_head.multi_level_cls_convs, self._bbox_head.multi_level_conv_cls
                )
                cls_scores = list(map_results)
            else:
                raise NotImplementedError(
                    "Not supported detection head provided. "
                    "DetSaliencyMapHook supports only the following single stage detectors: "
                    "YOLOXHead, ATSSHead, SSDHead, VFNetHead."
                )
        return cls_scores


class ReciproCAMHook(BaseRecordingForwardHook):
    """
    Implementation of recipro-cam for class-wise saliency map
    recipro-cam: gradient-free reciprocal class activation map (https://arxiv.org/pdf/2209.14074.pdf)
    """

    def __init__(self, module: torch.nn.Module, fpn_idx: int = -1) -> None:
        super().__init__(module, fpn_idx)
        self._neck = module.neck if module.with_neck else None
        self._head = module.head
        self._num_classes = module.head.num_classes

    def func(self, feature_map: Union[torch.Tensor, Sequence[torch.Tensor]], fpn_idx: int = -1) -> torch.Tensor:
        """
        Generate the class-wise saliency maps using Recipro-CAM and then normalizing to (0, 255).

        Args:
            feature_map (Union[torch.Tensor, List[torch.Tensor]]): feature maps from backbone or list of feature maps
                                                                    from FPN.
            fpn_idx (int, optional): The layer index to be processed if the model is a FPN.
                                      Defaults to 0 which uses the largest feature map from FPN.

        Returns:
            torch.Tensor: Class-wise Saliency Maps. One saliency map per each class - [batch, class_id, H, W]
        """
        if isinstance(feature_map, (list, tuple)):
            feature_map = feature_map[fpn_idx]

        bs, c, h, w = feature_map.size()
        saliency_maps = torch.empty(bs, self._num_classes, h, w)
        for f in range(bs):
            mosaic_feature_map = self._get_mosaic_feature_map(feature_map[f], c, h, w)
            mosaic_prediction = self._predict_from_feature_map(mosaic_feature_map)
            saliency_maps[f] = mosaic_prediction.transpose(0, 1).reshape((self._num_classes, h, w))

        saliency_maps = saliency_maps.reshape((bs, self._num_classes, h * w))
        max_values, _ = torch.max(saliency_maps, -1)
        min_values, _ = torch.min(saliency_maps, -1)
        saliency_maps = 255 * (saliency_maps - min_values[:, :, None]) / (max_values - min_values + 1e-12)[:, :, None]
        saliency_maps = saliency_maps.reshape((bs, self._num_classes, h, w))
        saliency_maps = saliency_maps.to(torch.uint8)
        return saliency_maps

    def _predict_from_feature_map(self, x: torch.Tensor) -> torch.Tensor:
        with torch.no_grad():
            if self._neck is not None:
                x = self._neck(x)
            logits = self._head.simple_test(x)
            if not isinstance(logits, torch.Tensor):
                logits = torch.tensor(logits)
        return logits

    def _get_mosaic_feature_map(self, feature_map: torch.Tensor, c: int, h: int, w: int) -> torch.Tensor:
        if self._neck is not None and isinstance(self._neck, GlobalAveragePooling):
            """
            Optimization workaround for the GAP case (simulate GAP with more simple compute graph)
            Possible due to static sparsity of mosaic_feature_map
            Makes the downstream GAP operation to be dummy
            """
            feature_map_transposed = torch.flatten(feature_map, start_dim=1).transpose(0, 1)[:, :, None, None]
            mosaic_feature_map = feature_map_transposed / (h * w)
        else:
            feature_map_repeated = feature_map.repeat(h * w, 1, 1, 1)
            mosaic_feature_map_mask = torch.zeros(h * w, c, h, w).to(feature_map.device)
            spacial_order = torch.arange(h * w).reshape(h, w)
            for i in range(h):
                for j in range(w):
                    k = spacial_order[i, j]
                    mosaic_feature_map_mask[k, :, i, j] = torch.ones(c).to(feature_map.device)
            mosaic_feature_map = feature_map_repeated * mosaic_feature_map_mask
        return mosaic_feature_map



"""EfficientNet-V2 for multi-class config."""

# pylint: disable=invalid-name

_base_ = "../base/models/efficientnet_v2.py"

model = dict(
    type="SAMImageClassifier",
    task="classification",
    backbone=dict(
        version="s_21k",
    ),
    head=dict(type="CustomLinearClsHead", loss=dict(type="CrossEntropyLoss", loss_weight=1.0)),
)

fp16 = dict(loss_scale=512.0)



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .mmov_fpn import MMOVFPN
from .mmov_ssd_neck import MMOVSSDNeck

# flake8: noqa
from .mmov_yolov3_neck import MMOVYOLOV3Neck



"""Data Pipeline for Cls-Incr model of Segmentation Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

# pylint: disable=invalid-name

__img_norm_cfg = dict(mean=[123.675, 116.28, 103.53], std=[58.395, 57.12, 57.375], to_rgb=True)
__img_scale = (544, 544)
__crop_size = (512, 512)

train_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(type="LoadAnnotations"),
    dict(type="Resize", img_scale=__img_scale, ratio_range=(0.5, 2.0), keep_ratio=False),
    dict(type="RandomCrop", crop_size=__crop_size, cat_max_ratio=0.75),
    dict(type="RandomFlip", prob=0.5, direction="horizontal"),
    dict(
        type="MaskCompose",
        prob=0.5,
        lambda_limits=(4, 16),
        keep_original=False,
        transforms=[
            dict(type="PhotoMetricDistortion"),
        ],
    ),
    dict(type="Normalize", **__img_norm_cfg),
    dict(type="Pad", size=__crop_size, pad_val=0, seg_pad_val=255),
    dict(type="RandomRotate", prob=0.5, degree=30, pad_val=0, seg_pad_val=255),
    dict(type="DefaultFormatBundle"),
    dict(type="Collect", keys=["img", "gt_semantic_seg"]),
]
test_pipeline = [
    dict(type="LoadImageFromFile"),
    dict(
        type="MultiScaleFlipAug",
        img_scale=__img_scale,
        flip=False,
        transforms=[
            dict(type="Resize", keep_ratio=False),
            dict(type="RandomFlip"),
            dict(type="Normalize", **__img_norm_cfg),
            dict(type="ImageToTensor", keys=["img"]),
            dict(type="Collect", keys=["img"]),
        ],
    ),
]

# TODO (Soobee) : Remove Repeatdataset in data config
data = dict(
    train=dict(dataset=dict(pipeline=train_pipeline)),
    val=dict(pipeline=test_pipeline),
    test=dict(pipeline=test_pipeline),
)



"""OTX Adapters for action recognition backbones - mmaction2."""

# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0


from .roi_head import AVARoIHead

__all__ = ["AVARoIHead"]



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import math

from mmcv.runner import HOOKS, Hook, LrUpdaterHook
from mmcv.runner.hooks.checkpoint import CheckpointHook
from mmcv.runner.hooks.evaluation import EvalHook

from otx.mpa.modules.hooks.early_stopping_hook import EarlyStoppingHook
from otx.mpa.utils.logger import get_logger

logger = get_logger()


@HOOKS.register_module()
class AdaptiveTrainSchedulingHook(Hook):
    """Adaptive Training Scheduling Hook

    Depending on the size of iteration per epoch, adaptively update the validation interval and related values.

    Args:
        max_interval (int): Maximum value of validation interval.
            Defaults to 5.
        base_lr_patience (int): The value of LR drop patience are expected in total epoch.
            Patience used when interval is 1, Defaults to 5.
        min_lr_patience (int): Minumum value of LR drop patience.
            Defaults to 2.
        base_es_patience (int): The value of Early-Stopping patience are expected in total epoch.
            Patience used when interval is 1, Defaults to 10.
    """

    def __init__(
        self,
        max_interval=5,
        base_lr_patience=5,
        min_lr_patience=2,
        base_es_patience=10,
        min_es_patience=3,
        decay=-0.025,
        enable_adaptive_interval_hook=False,
        enable_eval_before_run=False,
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.max_interval = max_interval
        self.base_lr_patience = base_lr_patience
        self.min_lr_patience = min_lr_patience
        self.base_es_patience = base_es_patience
        self.min_es_patience = min_es_patience
        self.decay = decay
        self.enable_adaptive_interval_hook = enable_adaptive_interval_hook
        self.enable_eval_before_run = enable_eval_before_run

        self._initialized = False
        self._original_interval = None

    def before_run(self, runner):
        if self.enable_eval_before_run:
            hook = self.get_evalhook(runner)
            if hook is None:
                logger.warning("EvalHook is not found in runner. Skipping enabling evaluation before run.")
                return
            self._original_interval = hook.interval
            hook.interval = 1
            hook.start = 0

    def before_train_iter(self, runner):
        if self.enable_eval_before_run and self._original_interval is not None:
            hook = self.get_evalhook(runner)
            hook.interval = self._original_interval
            self._original_interval = None

        if self.enable_adaptive_interval_hook and not self._initialized:
            self.max_interval = min(self.max_interval, runner.max_epochs - runner.epoch)
            iter_per_epoch = len(runner.data_loader)
            adaptive_interval = self.get_adaptive_interval(iter_per_epoch)
            for hook in runner.hooks:
                if isinstance(hook, EvalHook):
                    # make sure evaluation is done at last to save best checkpoint
                    limit = runner.max_epochs if hook.by_epoch else runner.max_iters
                    adaptive_interval = min(adaptive_interval, limit)
                    logger.info(f"Update EvalHook interval: {hook.interval} -> {adaptive_interval}")
                    hook.interval = adaptive_interval
                elif isinstance(hook, LrUpdaterHook):
                    patience = max(
                        math.ceil((self.base_lr_patience / adaptive_interval)),
                        self.min_lr_patience,
                    )
                    logger.info(f"Update LrUpdaterHook patience: {hook.patience} -> {patience}")
                    hook.interval = adaptive_interval
                    hook.patience = patience
                elif isinstance(hook, EarlyStoppingHook):
                    patience = max(
                        math.ceil((self.base_es_patience / adaptive_interval)),
                        self.min_es_patience,
                    )
                    logger.info(f"Update EarlyStoppingHook patience: {hook.patience} -> {patience}")
                    hook.start = adaptive_interval
                    hook.interval = adaptive_interval
                    hook.patience = patience
                elif isinstance(hook, CheckpointHook):
                    # make sure checkpoint is saved at last
                    limit = runner.max_epochs if hook.by_epoch else runner.max_iters
                    adaptive_interval = min(adaptive_interval, limit)
                    logger.info(f"Update CheckpointHook interval: {hook.interval} -> {adaptive_interval}")
                    hook.interval = adaptive_interval
            self._initialized = True

    def get_adaptive_interval(self, iter_per_epoch):
        adaptive_interval = max(round(math.exp(self.decay * iter_per_epoch) * self.max_interval), 1)
        return adaptive_interval

    def get_evalhook(self, runner):
        target_hook = None
        for hook in runner.hooks:
            if isinstance(hook, EvalHook):
                assert target_hook is None, "More than 1 EvalHook is found in runner."
                target_hook = hook
        return target_hook



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch.nn as nn
import torch.nn.functional as F
from mmcls.models.builder import HEADS
from mmcls.models.heads import ClsHead


@HEADS.register_module()
class ConvClsHead(ClsHead):
    """Convolutional classifier head.

    Args:
        num_classes (int): Number of categories excluding the background
            category.
        in_channels (int): Number of channels in the input feature map.
        init_cfg (dict | optional): The extra init config of layers.
            Defaults to use dict(type='Normal', layer='Linear', std=0.01).
    """

    def __init__(self, num_classes, in_channels, init_cfg=dict(type="Kaiming", layer=["Conv2d"]), *args, **kwargs):
        super(ConvClsHead, self).__init__(init_cfg=init_cfg, *args, **kwargs)

        self.in_channels = in_channels
        self.num_classes = num_classes

        if self.num_classes <= 0:
            raise ValueError(f"num_classes={num_classes} must be a positive integer")

        self.conv = nn.Conv2d(self.in_channels, self.num_classes, (1, 1))

    def pre_logits(self, x):
        if isinstance(x, tuple):
            x = x[-1]
        return x

    def simple_test(self, x, softmax=True, post_process=True):
        """Inference without augmentation.

        Args:
            x (tuple[Tensor]): The input features.
                Multi-stage inputs are acceptable but only the last stage will
                be used to classify. The shape of every item should be
                ``(num_samples, in_channels)``.
            softmax (bool): Whether to softmax the classification score.
            post_process (bool): Whether to do post processing the
                inference results. It will convert the output to a list.

        Returns:
            Tensor | list: The inference results.

                - If no post processing, the output is a tensor with shape
                  ``(num_samples, num_classes)``.
                - If post processing, the output is a multi-dimentional list of
                  float and the dimensions are ``(num_samples, num_classes)``.
        """
        x = self.pre_logits(x)
        cls_score = self.conv(x).squeeze()

        if softmax:
            pred = F.softmax(cls_score, dim=1) if cls_score is not None else None
        else:
            pred = cls_score

        if post_process:
            return self.post_process(pred)
        else:
            return pred

    def forward_train(self, x, gt_label, **kwargs):
        x = self.pre_logits(x)
        cls_score = self.conv(x).squeeze()
        losses = self.loss(cls_score, gt_label, **kwargs)
        return losses



"""Inference Task of OTX Segmentation."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import os
from typing import Dict, Optional

import numpy as np
from mmcv.utils import ConfigDict

from otx.algorithms.common.adapters.mmcv.utils import (
    patch_data_pipeline,
    patch_default_config,
    patch_runner,
    remove_from_configs_by_type,
)
from otx.algorithms.common.configs import TrainType
from otx.algorithms.common.tasks import BaseTask
from otx.algorithms.common.utils.callback import InferenceProgressCallback
from otx.algorithms.segmentation.adapters.mmseg.utils.builder import build_segmentor
from otx.algorithms.segmentation.adapters.mmseg.utils.config_utils import (
    patch_datasets,
    patch_evaluation,
)
from otx.algorithms.segmentation.adapters.openvino.model_wrappers.blur import (
    get_activation_map,
)
from otx.algorithms.segmentation.configs.base import SegmentationConfig
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.inference_parameters import (
    default_progress_callback as default_infer_progress_callback,
)
from otx.api.entities.model import (
    ModelEntity,
    ModelFormat,
    ModelOptimizationType,
    ModelPrecision,
)
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.tensor import TensorEntity
from otx.api.serialization.label_mapper import label_schema_to_bytes
from otx.api.usecases.evaluation.metrics_helper import MetricsHelper
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.export_interface import ExportType, IExportTask
from otx.api.usecases.tasks.interfaces.inference_interface import IInferenceTask
from otx.api.usecases.tasks.interfaces.unload_interface import IUnload
from otx.api.utils.argument_checks import (
    DatasetParamTypeCheck,
    check_input_parameters_type,
)
from otx.api.utils.segmentation_utils import (
    create_annotation_from_segmentation_map,
    create_hard_prediction_from_soft_prediction,
)
from otx.mpa import MPAConstants
from otx.mpa.utils.config_utils import MPAConfig
from otx.mpa.utils.logger import get_logger

logger = get_logger()


RECIPE_TRAIN_TYPE = {
    TrainType.SEMISUPERVISED: "semisl.py",
    TrainType.INCREMENTAL: "incremental.py",
    TrainType.SELFSUPERVISED: "selfsl.py",
}


# pylint: disable=too-many-locals, too-many-instance-attributes, attribute-defined-outside-init
class SegmentationInferenceTask(BaseTask, IInferenceTask, IExportTask, IEvaluationTask, IUnload):
    """Inference Task Implementation of OTX Segmentation."""

    @check_input_parameters_type()
    def __init__(self, task_environment: TaskEnvironment, **kwargs):
        # self._should_stop = False
        self.freeze = True
        self.metric = "mDice"
        self._label_dictionary = {}  # type: Dict

        super().__init__(SegmentationConfig, task_environment, **kwargs)
        self._label_dictionary = dict(enumerate(self._labels, 1))

    @check_input_parameters_type({"dataset": DatasetParamTypeCheck})
    def infer(
        self, dataset: DatasetEntity, inference_parameters: Optional[InferenceParameters] = None
    ) -> DatasetEntity:
        """Main infer function of OTX Segmentation."""
        logger.info("infer()")

        if inference_parameters is not None:
            update_progress_callback = inference_parameters.update_progress
            is_evaluation = inference_parameters.is_evaluation
        else:
            update_progress_callback = default_infer_progress_callback
            is_evaluation = False

        self._time_monitor = InferenceProgressCallback(len(dataset), update_progress_callback)

        stage_module = "SegInferrer"
        self._data_cfg = self._init_test_data_cfg(dataset)

        dump_features = True

        results = self._run_task(
            stage_module,
            mode="train",
            dataset=dataset,
            dump_features=dump_features,
        )
        logger.debug(f"result of run_task {stage_module} module = {results}")
        predictions = results["outputs"]
        prediction_results = zip(predictions["eval_predictions"], predictions["feature_vectors"])
        self._add_predictions_to_dataset(prediction_results, dataset, dump_soft_prediction=not is_evaluation)
        return dataset

    @check_input_parameters_type()
    def evaluate(self, output_resultset: ResultSetEntity, evaluation_metric: Optional[str] = None):
        """Evaluate function of OTX Segmentation Task."""
        logger.info("called evaluate()")

        if evaluation_metric is not None:
            logger.warning(
                f"Requested to use {evaluation_metric} metric, " "but parameter is ignored. Use mDice instead."
            )
        logger.info("Computing mDice")
        metrics = MetricsHelper.compute_dice_averaged_over_pixels(output_resultset)
        logger.info(f"mDice after evaluation: {metrics.overall_dice.value}")
        output_resultset.performance = metrics.get_performance()

    def unload(self):
        """Unload the task."""
        self.cleanup()

    @check_input_parameters_type()
    def export(self, export_type: ExportType, output_model: ModelEntity, dump_features: bool = True):
        """Export function of OTX Segmentation Task."""
        logger.info("Exporting the model")
        if export_type != ExportType.OPENVINO:
            raise RuntimeError(f"not supported export type {export_type}")
        output_model.model_format = ModelFormat.OPENVINO
        output_model.optimization_type = ModelOptimizationType.MO
        # TODO: add dumping saliency maps and representation vectors according to dump_features flag
        if not dump_features:
            logger.warning(
                "Ommitting feature dumping is not implemented."
                "The saliency maps and representation vector outputs will be dumped in the exported model."
            )

        stage_module = "SegExporter"
        results = self._run_task(
            stage_module,
            mode="train",
            export=True,
            dump_features=dump_features,
        )
        outputs = results.get("outputs")
        logger.debug(f"results of run_task = {outputs}")
        if outputs is None:
            raise RuntimeError(results.get("msg"))

        bin_file = outputs.get("bin")
        xml_file = outputs.get("xml")
        if xml_file is None or bin_file is None:
            raise RuntimeError("invalid status of exporting. bin and xml should not be None")
        with open(bin_file, "rb") as f:
            output_model.set_data("openvino.bin", f.read())
        with open(xml_file, "rb") as f:
            output_model.set_data("openvino.xml", f.read())
        output_model.precision = [ModelPrecision.FP32]
        output_model.optimization_methods = self._optimization_methods
        output_model.set_data("label_schema.json", label_schema_to_bytes(self._task_environment.label_schema))
        logger.info("Exporting completed")

    def _init_recipe(self):
        logger.info("called _init_recipe()")

        recipe_root = os.path.join(MPAConstants.RECIPES_PATH, "stages/segmentation")
        logger.info(f"train type = {self._train_type}")

        if self._train_type in RECIPE_TRAIN_TYPE:
            if self._train_type == TrainType.INCREMENTAL and self._hyperparams.learning_parameters.enable_supcon:
                recipe = os.path.join(recipe_root, "supcon.py")
                if "supcon" not in self._model_dir:
                    self._model_dir = os.path.join(self._model_dir, "supcon")
            else:
                recipe = os.path.join(recipe_root, RECIPE_TRAIN_TYPE[self._train_type])
        else:
            raise NotImplementedError(f"Train type {self._train_type} is not implemented yet.")

        logger.info(f"train type = {self._train_type} - loading {recipe}")

        self._recipe_cfg = MPAConfig.fromfile(recipe)
        options_for_patch_datasets = {"type": "MPASegDataset"}
        patch_default_config(self._recipe_cfg)
        patch_runner(self._recipe_cfg)
        patch_data_pipeline(self._recipe_cfg, self.data_pipeline_path)
        patch_datasets(
            self._recipe_cfg,
            self._task_type.domain,
            **options_for_patch_datasets,
        )  # for OTX compatibility
        patch_evaluation(self._recipe_cfg)  # for OTX compatibility
        if self._recipe_cfg.get("evaluation", None):
            self.metric = self._recipe_cfg.evaluation.metric

        if self._recipe_cfg.get("override_configs", None):
            self.override_configs.update(self._recipe_cfg.override_configs)

        if not self.freeze:
            remove_from_configs_by_type(self._recipe_cfg.custom_hooks, "FreezeLayers")
        logger.info(f"initialized recipe = {recipe}")

    def _update_stage_module(self, stage_module: str):
        module_prefix = {TrainType.SEMISUPERVISED: "SemiSL", TrainType.INCREMENTAL: "Incr"}
        if self._train_type == TrainType.SEMISUPERVISED and stage_module == "SegExporter":
            stage_module = "SemiSLSegExporter"
        elif self._train_type in module_prefix and stage_module in ["SegTrainer", "SegInferrer"]:
            stage_module = module_prefix[self._train_type] + stage_module

        return stage_module

    def _init_model_cfg(self):
        model_cfg = MPAConfig.fromfile(os.path.join(self._model_dir, "model.py"))
        return model_cfg

    def _init_test_data_cfg(self, dataset: DatasetEntity):
        data_cfg = ConfigDict(
            data=ConfigDict(
                train=ConfigDict(
                    otx_dataset=None,
                    labels=self._labels,
                ),
                test=ConfigDict(
                    otx_dataset=dataset,
                    labels=self._labels,
                ),
            )
        )
        return data_cfg

    def _add_predictions_to_dataset(self, prediction_results, dataset, dump_soft_prediction):
        """Loop over dataset again to assign predictions. Convert from MMSegmentation format to OTX format."""

        for dataset_item, (prediction, feature_vector) in zip(dataset, prediction_results):
            soft_prediction = np.transpose(prediction[0], axes=(1, 2, 0))
            hard_prediction = create_hard_prediction_from_soft_prediction(
                soft_prediction=soft_prediction,
                soft_threshold=self._hyperparams.postprocessing.soft_threshold,
                blur_strength=self._hyperparams.postprocessing.blur_strength,
            )
            annotations = create_annotation_from_segmentation_map(
                hard_prediction=hard_prediction,
                soft_prediction=soft_prediction,
                label_map=self._label_dictionary,
            )
            dataset_item.append_annotations(annotations=annotations)

            if feature_vector is not None:
                active_score = TensorEntity(name="representation_vector", numpy=feature_vector.reshape(-1))
                dataset_item.append_metadata_item(active_score, model=self._task_environment.model)

            if dump_soft_prediction:
                for label_index, label in self._label_dictionary.items():
                    if label_index == 0:
                        continue
                    current_label_soft_prediction = soft_prediction[:, :, label_index]
                    class_act_map = get_activation_map(current_label_soft_prediction)
                    result_media = ResultMediaEntity(
                        name=label.name,
                        type="soft_prediction",
                        label=label,
                        annotation_scene=dataset_item.annotation_scene,
                        roi=dataset_item.roi,
                        numpy=class_act_map,
                    )
                    dataset_item.append_metadata_item(result_media, model=self._task_environment.model)

    def _initialize_post_hook(self, options=None):
        super()._initialize_post_hook(options)
        options["model_builder"] = build_segmentor



"""This module contains the definition for the `convert` function within the configuration helper.

This function can be used to convert a OTX configuration object to a dictionary or yaml representation.
"""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from enum import Enum
from typing import Type, TypeVar

import yaml
from omegaconf import DictConfig, OmegaConf

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.elements import (
    ConfigurableEnum,
    ParameterGroup,
    metadata_keys,
)

ConvertTypeVar = TypeVar("ConvertTypeVar", str, DictConfig, dict)


def serialize_metadata(metadata_dict: dict, enum_to_str: bool = True) -> dict:
    """This function converts Enums in the metadata_dict to their string representation.

    It is used when converting between yaml and python object representation of the configuration.

    Args:
        metadata_dict (dict): Dictionary containing the metadata to convert
        enum_to_str (bool): True to convert enums to their string representation, False to leave them as Enums
    """
    for key, value in metadata_dict.items():
        if isinstance(value, Enum):
            if enum_to_str:
                metadata_dict[key] = str(value)
        if key == metadata_keys.UI_RULES:
            metadata_dict[key] = value.to_dict(enum_to_str)
    return metadata_dict


def parameter_group_to_dict(
    parameter_group: ParameterGroup,
    enum_to_str: bool = False,
    values_only: bool = False,
) -> dict:
    """Converts an instance of a `ParameterGroup` configuration element to its dictionary representation.

    Args:
        parameter_group (ParameterGroup): ParameterGroup to convert to dictionary representation
        enum_to_str (bool): Set to True to convert any Enum fields in the configuration to
            their string representation.
        values_only (bool): True to keep only the parameter values, and remove all meta
            data from the output dictionary

    Returns:
        dict: Nested dictionary with keys and values corresponding to the configuration
            defined in the instance of `ParameterGroup` for which the `to_dict` method was
            called.
    """
    parameter_group.update_auto_hpo_states()
    attribute_names = [attribute.name for attribute in parameter_group.__attrs_attrs__]  # type: ignore
    # The __attrs_attrs__ attribute is added through the attrs package, mypy doesn't recognize it so we can ignore the
    # type error
    attribute_values = [getattr(parameter_group, attribute_name) for attribute_name in attribute_names]

    dictionary_representation = {}
    for name, value in zip(attribute_names, attribute_values):
        # Go through all simple attributes first, add them to the dictionary representation
        if isinstance(value, Enum) and enum_to_str:
            value = str(value)
        dictionary_representation[name] = value
    for group_name in parameter_group.groups:
        # Then, recursively add all parameter groups to the dictionary representation
        group = getattr(parameter_group, group_name)
        dictionary_representation.update({group_name: parameter_group_to_dict(group, enum_to_str, values_only)})
    for parameter_name in parameter_group.parameters:
        # Then, add all parameters for this group to the dictionary representation
        # For each parameter, construct a dict with its value and then update it with the metadata
        value = getattr(parameter_group, parameter_name)
        if isinstance(value, ConfigurableEnum) and enum_to_str:
            value = str(value)
        if values_only:
            parameter_dictionary = value
        else:
            parameter_dictionary = {"value": value}
            parameter_dictionary.update(serialize_metadata(parameter_group.get_metadata(parameter_name), enum_to_str))
        # Finally, add the parameter to the dictionary representation of the group
        dictionary_representation.update({parameter_name: parameter_dictionary})
    return dictionary_representation


def convert(
    config: ConfigurableParameters,
    target: Type[ConvertTypeVar],
    enum_to_str: bool = False,
    id_to_str: bool = False,
    values_only: bool = False,
) -> ConvertTypeVar:
    """Convert a configuration object to either a yaml string, a dictionary or an OmegaConf DictConfig object.

    Args:
        config (ConfigurableParameters): ConfigurableParameters object to convert
        target (Type[ConvertTypeVar]): target type to convert to. Options are [str, dict, DictConfig]
        enum_to_str (bool) : Boolean specifying whether to convert enums within the config
                        to their string representation. For conversion to yaml, enums
                        are automatically converted and this option is disregarded.
        id_to_str (bool): True to convert the id of the configurable parameters to a string
            representation, False to leave it as an ID object
        values_only (bool): True to keep only the parameter values, and remove all meta
            data from the target output

    Raises:
        ValueError: if an unsupported conversion target is supplied

    Returns:
        ConvertTypeVar: Result of the conversion, the configuration specified in `config` in the
            representation specified in `target`
    """
    if target == str:
        enum_to_str = True

    config_dict = parameter_group_to_dict(config, enum_to_str=enum_to_str, values_only=values_only)

    if id_to_str or target == str or target == DictConfig:
        config_id = config_dict.get("id", None)
        config_dict["id"] = str(config_id) if config_id is not None else None

    if target == str:
        result = yaml.dump(config_dict)
    elif target == dict:
        result = config_dict
    elif target == DictConfig:
        result = OmegaConf.create(config_dict)
    else:
        raise ValueError("Unsupported conversion target! Supported target types are [str, dict, DictConfig]")
    return result  # type: ignore



"""This module implements the Rectangle shape entity."""
# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

# Conflict with Isort
# pylint: disable=wrong-import-order, cyclic-import

import datetime
import math
import warnings
from typing import Optional

import numpy as np
from shapely.geometry import Polygon as shapely_polygon

from otx.api.entities.shapes.shape import Shape, ShapeEntity, ShapeType
from otx.api.utils.time_utils import now

# pylint: disable=invalid-name


class Rectangle(Shape):
    """Rectangle represents a rectangular shape.

    Rectangle are used to annotate detection and classification tasks. In the
    classification case, the rectangle is a full rectangle spanning the whole related
    item (could be an image, video frame, a region of interest).

    - x1 and y1 represent the top-left coordinate of the rectangle
    - x2 and y2 representing the bottom-right coordinate of the rectangle

    Args:
        x1 (float): x-coordinate of the top-left corner of the rectangle
        y1 (float): y-coordinate of the top-left corner of the rectangle
        x2 (float): x-coordinate of the bottom-right corner of the rectangle
        y2 (float): y-coordinate of the bottom-right corner of the rectangle
        modification_date (datetime.datetime): Date of the last modification of the rectangle
    """

    # pylint: disable=too-many-arguments; Requires refactor
    def __init__(
        self,
        x1: float,
        y1: float,
        x2: float,
        y2: float,
        modification_date: Optional[datetime.datetime] = None,
    ):
        modification_date = now() if modification_date is None else modification_date
        super().__init__(
            shape_type=ShapeType.RECTANGLE,
            modification_date=modification_date,
        )

        is_valid = True
        for (x, y) in [(x1, y1), (x2, y2)]:
            is_valid = is_valid and self._validate_coordinates(x, y)
        if not is_valid:
            warnings.warn(
                f"{type(self).__name__} coordinates are invalid : x1={x1}, y1={y1}, x2={x2}, y2={y2}",
                UserWarning,
            )

        self.x1 = x1
        self.y1 = y1
        self.x2 = x2
        self.y2 = y2
        if self.width <= 0 or self.height <= 0:
            raise ValueError(
                f"Invalid rectangle with coordinates: x1={self.x1}, y1={self.y1}, " f"x2={self.x2}, y2={self.y2}"
            )

    def __repr__(self):
        """String representation of the rectangle."""
        return f"Rectangle(x={self.x1}, y={self.y1}, width={self.width}, " f"height={self.height})"

    def __eq__(self, other: object):
        """Returns True if `other` is a `Rectangle` with the same coordinates."""
        if isinstance(other, Rectangle):
            return (
                self.x1 == other.x1
                and self.y1 == other.y1
                and self.x2 == other.x2
                and self.y2 == other.y2
                and self.modification_date == other.modification_date
            )
        return False

    def __hash__(self):
        """Returns hash of the rectangle."""
        return hash(str(self))

    def clip_to_visible_region(self) -> "Rectangle":
        """Clip the rectangle to the [0, 1] visible region of an image.

        Returns:
            Rectangle: Clipped rectangle.
        """
        x1 = min(max(0.0, self.x1), 1.0)
        y1 = min(max(0.0, self.y1), 1.0)
        x2 = min(max(0.0, self.x2), 1.0)
        y2 = min(max(0.0, self.y2), 1.0)

        return Rectangle(x1=x1, y1=y1, x2=x2, y2=y2, modification_date=self.modification_date)

    def normalize_wrt_roi_shape(self, roi_shape: ShapeEntity) -> "Rectangle":
        """Transforms from the `roi` coordinate system to the normalized coordinate system.

        Example:
            Assume we have rectangle `b1` which lives in the top-right quarter of
            a 2D space. The 2D space where `b1` lives in is an `roi` living in the top-left
            quarter of the normalized coordinate space. This function returns rectangle
            `b1` expressed in the normalized coordinate space.

                >>> from otx.api.entities.annotation import Annotation
                >>> b1 = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=0.5)
                >>> roi = Rectangle(x1=0.0, x2=0.5, y1=0.0, y2=0.5)
                >>> normalized = b1.normalize_wrt_roi_shape(roi_shape)
                >>> normalized
                Box(, x=0.25, y=0.0, width=0.25, height=0.25)

        Args:
            roi_shape (ShapeEntity): Region of Interest.

        Raises:
            ValueError: If the `roi_shape` is not a `Rectangle`.

        Returns:
            New polygon in the image coordinate system
        """
        if not isinstance(roi_shape, Rectangle):
            raise ValueError("roi_shape has to be a Rectangle.")

        roi_shape = roi_shape.clip_to_visible_region()

        return Rectangle(
            x1=self.x1 * roi_shape.width + roi_shape.x1,
            y1=self.y1 * roi_shape.height + roi_shape.y1,
            x2=self.x2 * roi_shape.width + roi_shape.x1,
            y2=self.y2 * roi_shape.height + roi_shape.y1,
            modification_date=self.modification_date,
        )

    def denormalize_wrt_roi_shape(self, roi_shape: ShapeEntity) -> "Rectangle":
        """Transforming shape from the normalized coordinate system to the `roi` coordinate system.

        Example:

            Assume we have rectangle `b1` which lives in the top-right quarter of
            the normalized coordinate space. The `roi` is a rectangle living in the half
            right of the normalized coordinate space. This function returns rectangle
            `b1` expressed in the coordinate space of `roi`. (should return top-half)
            Box denormalized to a rectangle as ROI

                >>> from otx.api.entities.annotation import Annotation
                >>> b1 = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=0.5)
                # the top-right
                >>> roi = Annotation(Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=1.0))
                # the half-right
                >>> normalized = b1.denormalize_wrt_roi_shape(roi_shape)
                # should return top half
                >>> normalized
                Box(, x=0.0, y=0.0, width=1.0, height=0.5)

        Args:
            roi_shape (ShapeEntity): Region of Interest

        Raises:
            ValueError: If the `roi_shape` is not a `Rectangle`.

        Returns:
            Rectangle: New polygon in the ROI coordinate system
        """
        if not isinstance(roi_shape, Rectangle):
            raise ValueError("roi_shape has to be a Rectangle.")

        roi_shape = roi_shape.clip_to_visible_region()

        x1 = (self.x1 - roi_shape.x1) / roi_shape.width
        y1 = (self.y1 - roi_shape.y1) / roi_shape.height
        x2 = (self.x2 - roi_shape.x1) / roi_shape.width
        y2 = (self.y2 - roi_shape.y1) / roi_shape.height

        return Rectangle(
            x1=x1,
            y1=y1,
            x2=x2,
            y2=y2,
            modification_date=self.modification_date,
        )

    def _as_shapely_polygon(self) -> shapely_polygon:
        points = [
            (self.x1, self.y1),
            (self.x2, self.y1),
            (self.x2, self.y2),
            (self.x1, self.y2),
            (self.x1, self.y1),
        ]
        return shapely_polygon(points)

    @classmethod
    def generate_full_box(cls) -> "Rectangle":
        """Returns a rectangle that fully encapsulates the normalized coordinate space.

        Example:
            >>> Rectangle.generate_full_box()
            Box(, x=0.0, y=0.0, width=1.0, height=1.0)

        Returns:
            Rectangle: A rectangle that fully encapsulates the normalized coordinate space.
        """
        return cls(x1=0.0, y1=0.0, x2=1.0, y2=1.0)

    @staticmethod
    def is_full_box(rectangle: ShapeEntity) -> bool:
        """Returns true if rectangle is a full box (occupying the full normalized coordinate space).

        Example:

            >>> b1 = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=1.0)
            >>> Rectangle.is_full_box(b1)
            False

            >>> b2 = Rectangle(x1=0.0, x2=1.0, y1=0.0, y2=1.0)
            >>> Rectangle.is_full_box(b2)
            True

        Args:
            rectangle (ShapeEntity): rectangle to evaluate

        Returns:
            bool: true if it fully encapsulate normalized coordinate space.
        """
        if (
            isinstance(rectangle, Rectangle)
            and rectangle.x1 == 0
            and rectangle.y1 == 0
            and rectangle.height == 1
            and rectangle.width == 1
        ):
            return True
        return False

    def crop_numpy_array(self, data: np.ndarray) -> np.ndarray:
        """Crop the given Numpy array to the region of interest represented by this rectangle.

        Args:
            data (np.ndarray): Image to crop.

        Returns:
            np.ndarray: Cropped image.
        """

        # We clip negative values to zero since Numpy uses negative values
        # to represent indexing from the right side of the array.
        # However, on the other hand, it is safe to have indices larger than the size
        # of the dimension; therefore, we do not clip values larger than the width and
        # height.
        x1 = max(int(round(self.x1 * data.shape[1])), 0)
        x2 = max(int(round(self.x2 * data.shape[1])), 0)
        y1 = max(int(round(self.y1 * data.shape[0])), 0)
        y2 = max(int(round(self.y2 * data.shape[0])), 0)

        return data[y1:y2, x1:x2, ::]

    @property
    def width(self) -> float:
        """Returns the width of the rectangle (x-axis).

        Example:

            >>> b1 = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=0.5)
            >>> b1.width
            0.5

        Returns:
            float: the width of the rectangle. (x-axis)
        """
        return self.x2 - self.x1

    @property
    def height(self) -> float:
        """Returns the height of the rectangle (y-axis).

        Example:

            >>> b1 = Rectangle(x1=0.5, x2=1.0, y1=0.0, y2=0.5)
            >>> b1.height
            0.5

        Returns:
            float: the height of the rectangle. (y-axis)
        """
        return self.y2 - self.y1

    @property
    def diagonal(self) -> float:
        """Returns the diagonal size/hypotenuse  of the rectangle (x-axis).

        Example:

            >>> b1 = Rectangle(x1=0.0, x2=0.3, y1=0.0, y2=0.4)
            >>> b1.diagonal
            0.5

        Returns:
            float: the width of the rectangle. (x-axis)
        """
        return math.hypot(self.width, self.height)

    def get_area(self) -> float:
        """Computes the approximate area of the shape.

        Area is a value between 0 and 1, calculated as (x2-x1) * (y2-y1)

        NOTE: This method should not be relied on for exact area computation. The area
        is approximate, because shapes are continuous, but pixels are discrete.

        Example:
            >>> Rectangle(0, 0, 1, 1).get_area()
            1.0
            >>> Rectangle(0.5, 0.5, 1.0, 1.0).get_area()
            0.25

        Returns:
            float: Approximate area of the shape.
        """
        return (self.x2 - self.x1) * (self.y2 - self.y1)



"""Utils for HPO with hpopt."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import json
import logging
import re
import shutil
from copy import deepcopy
from enum import Enum
from functools import partial
from inspect import isclass
from math import floor
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import torch
import yaml

from otx.api.configuration.helper import create
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.model import ModelEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.subset import Subset
from otx.api.entities.task_environment import TaskEnvironment
from otx.api.entities.train_parameters import TrainParameters, UpdateProgressCallback
from otx.cli.utils.importing import get_impl_class
from otx.cli.utils.io import read_model, save_model_data
from otx.core.data.adapter import get_dataset_adapter
from otx.hpo import HyperBand, TrialStatus, run_hpo_loop

logger = logging.getLogger(__name__)


def _check_hpo_enabled_task(task_type):
    return task_type in [
        TaskType.CLASSIFICATION,
        TaskType.DETECTION,
        TaskType.SEGMENTATION,
        TaskType.INSTANCE_SEGMENTATION,
        TaskType.ROTATED_DETECTION,
        TaskType.ANOMALY_CLASSIFICATION,
        TaskType.ANOMALY_DETECTION,
        TaskType.ANOMALY_SEGMENTATION,
    ]


class TaskManager:
    """Task utility class to give common interface from different task.

    Args:
        task_type (TaskType): otx task type
    """

    def __init__(self, task_type: TaskType):
        self._task_type = task_type

    @property
    def task_type(self):
        """Task_type property."""
        return self._task_type

    def is_mmcv_framework_task(self) -> bool:
        """Check task is run on mmcv.

        Returns:
            bool: whether task is run on mmcv
        """
        return self.is_cls_framework_task() or self.is_det_framework_task() or self.is_seg_framework_task()

    def is_cls_framework_task(self) -> bool:
        """Check that task is run on mmcls framework.

        Returns:
            bool: whether task is run on mmcls
        """
        return self._task_type == TaskType.CLASSIFICATION

    def is_det_framework_task(self) -> bool:
        """Check that task is one of a task run on mmdet framework.

        Returns:
            bool: whether task is run on mmdet
        """
        return self._task_type in [
            TaskType.DETECTION,
            TaskType.INSTANCE_SEGMENTATION,
            TaskType.ROTATED_DETECTION,
        ]

    def is_seg_framework_task(self) -> bool:
        """Check that task is run on mmseg framework.

        Returns:
            bool: whether tasks is run on mmseg
        """
        return self._task_type == TaskType.SEGMENTATION

    def is_anomaly_framework_task(self) -> bool:
        """Check taht task is run on anomalib.

        Returns:
            bool: whether task is run on anomalib
        """
        return self._task_type in [
            TaskType.ANOMALY_CLASSIFICATION,
            TaskType.ANOMALY_DETECTION,
            TaskType.ANOMALY_SEGMENTATION,
        ]

    def get_batch_size_name(self) -> str:
        """Give an proper batch size name depending on framework.

        Returns:
            str: batch size name
        """
        if self.is_mmcv_framework_task():
            batch_size_name = "learning_parameters.batch_size"
        elif self.is_anomaly_framework_task():
            batch_size_name = "learning_parameters.train_batch_size"
        else:
            raise RuntimeError(f"There is no information about {self._task_type} batch size name")

        return batch_size_name

    def get_epoch_name(self) -> str:
        """Give an proper epoch name depending on framework.

        Returns:
            str: epoch name
        """
        if self.is_mmcv_framework_task():
            epoch_name = "num_iters"
        elif self.is_anomaly_framework_task():
            epoch_name = "max_epochs"
        else:
            raise RuntimeError(f"There is no information about {self._task_type} epoch name")

        return epoch_name

    def copy_weight(self, src: Union[str, Path], det: Union[str, Path]):
        """Copy all model weights from work directory.

        Args:
            src (Union[str, Path]): path where model weights are saved
            det (Union[str, Path]): path to save model weights
        """
        src = Path(src)
        det = Path(det)
        if self.is_mmcv_framework_task():
            for weight_candidate in src.rglob("*epoch*.pth"):
                if not (weight_candidate.is_symlink() or (det / weight_candidate.name).exists()):
                    shutil.copy(weight_candidate, det)
        # TODO need to implement after anomaly task supports resume

    def get_latest_weight(self, workdir: Union[str, Path]) -> Optional[str]:
        """Get latest model weight from all weights.

        Args:
            workdir (Union[str, Path]): path where model weights are saved

        Returns:
            Optional[str]: latest model weight path. If not found, than return None value.
        """
        latest_weight = None
        workdir = Path(workdir)
        if self.is_mmcv_framework_task():
            pattern = re.compile(r"(\d+)\.pth")
            current_latest_epoch = -1
            latest_weight = None

            for weight_name in workdir.rglob("epoch_*.pth"):
                ret = pattern.search(str(weight_name))
                if ret is not None:
                    epoch = int(ret.group(1))
                    if current_latest_epoch < epoch:
                        current_latest_epoch = epoch
                        latest_weight = str(weight_name)
        # TODO need to implement after anomaly task supports resume

        return latest_weight


class TaskEnvironmentManager:
    """OTX environment utility class to set or get a value from environment class.

    Args:
        environment (TaskEnvironment): OTX task environment
    """

    def __init__(self, environment: TaskEnvironment):
        self._environment = environment
        self.task = TaskManager(environment.model_template.task_type)

    @property
    def environment(self):
        """Environment property."""
        return self._environment

    def get_task(self) -> TaskType:
        """Get task type of environment.

        Returns:
            TaskType: task type
        """
        return self._environment.model_template.task_type

    def get_model_template(self):
        """Get model template."""
        return self._environment.model_template

    def get_model_template_path(self) -> str:
        """Get model template path.

        Returns:
            str: path of model template
        """
        return self._environment.model_template.model_template_path

    def set_hyper_parameter_using_str_key(self, hyper_parameter: Dict[str, Any]):
        """Set hyper parameter to environment using string key hyper_parameter.

        Set hyper parameter to environment. Argument `hyper_parameter` is a dictionary which has string key.
        For example, hyper_parameter has a key "a.b.c", then value is set at env_hp.a.b.c.

        Args:
            hyper_parameter (Dict[str, Any]): hyper parameter to set which has a string format
        """
        env_hp = self._environment.get_hyper_parameters()  # type: ignore

        for param_key, param_val in hyper_parameter.items():
            splited_param_key = param_key.split(".")

            target = env_hp
            for val in splited_param_key[:-1]:
                target = getattr(target, val)
            setattr(target, splited_param_key[-1], param_val)

    def get_dict_type_hyper_parameter(self) -> Dict[str, Any]:
        """Get dictionary type hyper parmaeter of environment.

        Returns:
            Dict[str, Any]: dictionary type hyper parameter of environment
        """
        learning_parameters = self._environment.get_hyper_parameters().learning_parameters  # type: ignore
        learning_parameters = self._convert_parameter_group_to_dict(learning_parameters)
        hyper_parameter = {f"learning_parameters.{key}": val for key, val in learning_parameters.items()}
        return hyper_parameter

    def _convert_parameter_group_to_dict(self, parameter_group) -> Dict[str, Any]:
        """Convert parameter group to dictionary.

        Args:
            parameter_group : parameter gruop

        Returns:
            Dict[str, Any]: parameter group converted to dictionary
        """
        groups = getattr(parameter_group, "groups", None)
        parameters = getattr(parameter_group, "parameters", None)

        total_arr = []
        for val in [groups, parameters]:
            if val is not None:
                total_arr.extend(val)
        if not total_arr:
            return parameter_group

        ret = {}
        for key in total_arr:
            val = self._convert_parameter_group_to_dict(getattr(parameter_group, key))
            if not (isclass(val) or isinstance(val, Enum)):
                ret[key] = val

        return ret

    def get_max_epoch(self) -> int:
        """Get max epoch from environment.

        Returns:
            int: max epoch of environment
        """
        return getattr(
            self._environment.get_hyper_parameters().learning_parameters, self.task.get_epoch_name()  # type: ignore
        )

    def save_initial_weight(self, save_path: Union[Path, str]) -> bool:
        """Save an initial model weight.

        Args:
            save_path (Union[str, Path]): path to save initial model weight

        Returns:
            bool: whether model weight is saved successfully
        """
        save_path = Path(save_path)
        dir_path = save_path.parent
        if self._environment.model is None:
            # if task isn't anomaly, then save model weight during first trial
            if self.task.is_anomaly_framework_task():
                task = self.get_train_task()
                model = self.get_new_model_entity()
                task.save_model(model)
                save_model_data(model, str(dir_path))
                (dir_path / "weights.pth").rename(save_path)
                return True
        else:
            save_model_data(self._environment.model, str(dir_path))
            (dir_path / "weights.pth").rename(save_path)
            return True
        return False

    def get_train_task(self):
        """Get OTX train task instance.

        Returns:
           OTX task: OTX train task instance
        """
        impl_class = get_impl_class(self._environment.model_template.entrypoints.base)
        return impl_class(task_environment=self._environment)

    def get_batch_size_name(self) -> str:
        """Get proper batch size name depending on task.

        Returns:
            str: batch size name
        """
        return self.task.get_batch_size_name()

    def load_model_weight(self, model_weight_path: str, dataset: DatasetEntity):
        """Set model weight on environment to load the weight during training.

        Args:
            model_weight_path (str): model weight to load during training
            dataset (DatasetEntity): dataset for training a model
        """
        self._environment.model = read_model(self._environment.get_model_configuration(), model_weight_path, dataset)

    def resume_model_weight(self, model_weight_path: str, dataset: DatasetEntity):
        """Set model weight on environment to resume the weight during training.

        Args:
            model_weight_path (str): model weight to resume during training
            dataset (DatasetEntity): dataset for training a model
        """
        self.load_model_weight(model_weight_path, dataset)
        self._environment.model.model_adapters["resume"] = True  # type: ignore

    def get_new_model_entity(self, dataset=None) -> ModelEntity:
        """Get new model entity using environment.

        Args:
            dataset (Optional[DatasetEntity]): OTX dataset

        Returns:
            ModelEntity: new model entity
        """
        return ModelEntity(
            dataset,
            self._environment.get_model_configuration(),
        )

    def set_epoch(self, epoch: int):
        """Set epoch on environment.

        Args:
            epoch (int): epoch to set
        """
        hyper_parameter = {f"learning_parameters.{self.task.get_epoch_name()}": epoch}
        self.set_hyper_parameter_using_str_key(hyper_parameter)


class HpoRunner:
    """Class which is in charge of preparing and running HPO.

    Args:
        environment (TaskEnvironment): OTX environment
        train_dataset_size (int): train dataset size
        val_dataset_size (int): validation dataset size
        hpo_workdir (Union[str, Path]): work directory for HPO
        hpo_time_ratio (int, optional): time ratio to use for HPO compared to training time. Defaults to 4.
    """

    # pylint: disable=too-many-instance-attributes

    def __init__(
        self,
        environment: TaskEnvironment,
        train_dataset_size: int,
        val_dataset_size: int,
        hpo_workdir: Union[str, Path],
        hpo_time_ratio: int = 4,
    ):
        if train_dataset_size <= 0:
            raise ValueError(f"train_dataset_size should be bigger than 0. Your value is {train_dataset_size}")
        if val_dataset_size <= 0:
            raise ValueError(f"val_dataset_size should be bigger than 0. Your value is {val_dataset_size}")
        if hpo_time_ratio < 1:
            raise ValueError(f"hpo_time_ratio shouldn't be smaller than 1. Your value is {hpo_time_ratio}")

        self._environment = TaskEnvironmentManager(environment)
        self._hpo_workdir: Path = Path(hpo_workdir)
        self._hpo_time_ratio = hpo_time_ratio
        self._hpo_config: Dict = self._set_hpo_config()
        self._train_dataset_size = train_dataset_size
        self._val_dataset_size = val_dataset_size
        self._fixed_hp: Dict[str, Any] = {}
        self._initial_weight_name = "initial_weight.pth"

        self._align_batch_size_search_space_to_dataset_size()

    def _set_hpo_config(self):
        hpo_config_path = Path(self._environment.get_model_template_path()).parent / "hpo_config.yaml"
        with hpo_config_path.open("r") as f:
            hpopt_cfg = yaml.safe_load(f)

        return hpopt_cfg

    def _align_batch_size_search_space_to_dataset_size(self):
        batch_size_name = self._environment.get_batch_size_name()

        if batch_size_name in self._hpo_config["hp_space"]:
            if "range" in self._hpo_config["hp_space"][batch_size_name]:
                max_val = self._hpo_config["hp_space"][batch_size_name]["range"][1]
                min_val = self._hpo_config["hp_space"][batch_size_name]["range"][0]
                if max_val > self._train_dataset_size:
                    max_val = self._train_dataset_size
                    self._hpo_config["hp_space"][batch_size_name]["range"][1] = max_val
            else:
                max_val = self._hpo_config["hp_space"][batch_size_name]["max"]
                min_val = self._hpo_config["hp_space"][batch_size_name]["min"]

                if max_val > self._train_dataset_size:
                    max_val = self._train_dataset_size
                    self._hpo_config["hp_space"][batch_size_name]["max"] = max_val

            # If trainset size is lower than min batch size range,
            # fix batch size to trainset size
            if min_val >= max_val:
                logger.info(
                    "Train set size is equal or lower than batch size range. Batch size is fixed to train set size."
                )
                del self._hpo_config["hp_space"][batch_size_name]
                self._fixed_hp[batch_size_name] = self._train_dataset_size
                self._environment.set_hyper_parameter_using_str_key(self._fixed_hp)

    def run_hpo(self, train_func: Callable, data_roots: Dict[str, str]) -> Dict[str, Any]:
        """Run HPO and provides optimized hyper parameters.

        Args:
            train_func (Callable): training model function
            data_roots (Dict[str, str]): dataset path of each dataset type

        Returns:
            Dict[str, Any]: optimized hyper parameters
        """
        self._environment.save_initial_weight(self._get_initial_model_weight_path())
        hpo_algo = self._get_hpo_algo()
        resource_type = "gpu" if torch.cuda.is_available() else "cpu"
        run_hpo_loop(
            hpo_algo,
            partial(
                train_func,
                model_template=self._environment.get_model_template(),
                data_roots=data_roots,
                task_type=self._environment.get_task(),
                hpo_workdir=self._hpo_workdir,
                initial_weight_name=self._initial_weight_name,
                metric=self._hpo_config["metric"],
            ),
            resource_type,  # type: ignore
        )
        best_config = hpo_algo.get_best_config()
        self._restore_fixed_hp(best_config["config"])
        hpo_algo.print_result()

        return best_config

    def _restore_fixed_hp(self, hyper_parameter: Dict[str, Any]):
        for key, val in self._fixed_hp.items():
            hyper_parameter[key] = val

    def _get_hpo_algo(self):
        hpo_algo_type = self._hpo_config.get("search_algorithm", "asha")

        if hpo_algo_type == "asha":
            hpo_algo = self._prepare_asha()
        elif hpo_algo_type == "smbo":
            hpo_algo = self._prepare_smbo()
        else:
            raise ValueError(f"Supported HPO algorithms are asha and smbo. your value is {hpo_algo_type}.")

        return hpo_algo

    def _prepare_asha(self):
        args = {
            "search_space": self._hpo_config["hp_space"],
            "save_path": str(self._hpo_workdir),
            "maximum_resource": self._hpo_config.get("maximum_resource"),
            "minimum_resource": self._hpo_config.get("minimum_resource"),
            "mode": self._hpo_config.get("mode", "max"),
            "num_workers": 1,
            "num_full_iterations": self._environment.get_max_epoch(),
            "full_dataset_size": self._train_dataset_size,
            "non_pure_train_ratio": self._val_dataset_size / (self._train_dataset_size + self._val_dataset_size),
            "metric": self._hpo_config.get("metric", "mAP"),
            "expected_time_ratio": self._hpo_time_ratio,
            "prior_hyper_parameters": self._get_default_hyper_parameters(),
            "asynchronous_bracket": True,
            "asynchronous_sha": torch.cuda.device_count() != 1,
        }

        logger.debug(f"ASHA args = {args}")

        return HyperBand(**args)

    def _prepare_smbo(self):
        raise NotImplementedError

    def _get_default_hyper_parameters(self):
        default_hyper_parameters = {}
        hp_from_env = self._environment.get_dict_type_hyper_parameter()

        for key, val in hp_from_env.items():
            if key in self._hpo_config["hp_space"]:
                default_hyper_parameters[key] = val

        if not default_hyper_parameters:
            return None
        return default_hyper_parameters

    def _get_initial_model_weight_path(self):
        return self._hpo_workdir / self._initial_weight_name


def run_hpo(
    args, environment: TaskEnvironment, dataset: DatasetEntity, data_roots: Dict[str, str]
) -> Optional[TaskEnvironment]:
    """Run HPO and load optimized hyper parameter and best HPO model weight.

    Args:
        args: arguments passed to otx train
        environment (TaskEnvironment): otx task environment
        dataset (DatasetEntity): dataset to use for training
        data_roots (Dict[str, str]): dataset path of each dataset type
    """
    task_type = environment.model_template.task_type
    if not _check_hpo_enabled_task(task_type):
        logger.warning(
            "Currently supported task types are classification, detection, segmentation and anomaly"
            f"{task_type} is not supported yet."
        )
        return None

    hpo_save_path = (Path(args.save_model_to).parent / "hpo").absolute()
    hpo_runner = HpoRunner(
        environment,
        len(dataset.get_subset(Subset.TRAINING)),
        len(dataset.get_subset(Subset.VALIDATION)),
        hpo_save_path,
        args.hpo_time_ratio,
    )

    logger.info("started hyper-parameter optimization")
    best_config = hpo_runner.run_hpo(run_trial, data_roots)
    logger.info("completed hyper-parameter optimization")

    env_manager = TaskEnvironmentManager(environment)
    env_manager.set_hyper_parameter_using_str_key(best_config["config"])
    best_hpo_weight = get_best_hpo_weight(hpo_save_path, best_config["id"])
    if best_hpo_weight is None:
        logger.warning("Can not find the best HPO weight. Best HPO wegiht won't be used.")
    else:
        logger.debug(f"{best_hpo_weight} will be loaded as best HPO weight")
        env_manager.load_model_weight(best_hpo_weight, dataset)

    return env_manager.environment


def get_best_hpo_weight(hpo_dir: Union[str, Path], trial_id: Union[str, Path]) -> Optional[str]:
    """Get best model weight path of the HPO trial.

    Args:
        hpo_dir (Union[str, Path]): HPO work directory path
        trial_id (Union[str, Path]): trial id

    Returns:
        Optional[str]: best HPO model weight
    """
    hpo_dir = Path(hpo_dir)
    trial_output_files = list(hpo_dir.rglob(f"{trial_id}.json"))
    if not trial_output_files:
        return None
    trial_output_file = trial_output_files[0]

    with trial_output_file.open("r") as f:
        trial_output = json.load(f)

    best_epochs = []
    best_score = None
    for eph, score in trial_output["score"].items():
        if best_score is None:
            best_score = score
            best_epochs.append(eph)
        elif best_score < score:
            best_score = score
            best_epochs = [eph]
        elif best_score == score:
            best_epochs.append(eph)

    best_weight = None
    for best_epoch in best_epochs:
        best_weight_path = list(hpo_dir.glob(f"weight/{trial_id}/*epoch*{best_epoch}*"))
        if best_weight_path:
            best_weight = str(best_weight_path[0])

    return best_weight


class Trainer:
    """Class which prepares and trains a model given hyper parameters.

    Args:
        hp_config (Dict[str, Any]): hyper parameter to use on training
        report_func (Callable): function to report score
        model_template: model template
        data_roots (Dict[str, str]): dataset path of each dataset type
        task_type (TaskType): OTX task type
        hpo_workdir (Union[str, Path]): work directory for HPO
        initial_weight_name (str): initial model weight name for each trials to load
        metric (str): metric name
    """

    # pylint: disable=too-many-arguments, too-many-instance-attributes

    def __init__(
        self,
        hp_config: Dict[str, Any],
        report_func: Callable,
        model_template,
        data_roots: Dict[str, str],
        task_type: TaskType,
        hpo_workdir: Union[str, Path],
        initial_weight_name: str,
        metric: str,
    ):
        self._hp_config = hp_config
        self._report_func = report_func
        self._model_template = model_template
        self._data_roots = data_roots
        self._task = TaskManager(task_type)
        self._hpo_workdir: Path = Path(hpo_workdir)
        self._initial_weight_name = initial_weight_name
        self._metric = metric
        self._epoch = floor(self._hp_config["configuration"]["iterations"])
        del self._hp_config["configuration"]["iterations"]

    def run(self):
        """Run each training of each trial with given hyper parameters."""
        hyper_parameters = self._prepare_hyper_parameter()
        dataset_adapter = self._prepare_dataset_adapter()

        dataset = dataset_adapter.get_otx_dataset()
        dataset = HpoDataset(dataset, self._hp_config)

        label_schema = dataset_adapter.get_label_schema()

        environment = self._prepare_environment(hyper_parameters, label_schema)
        self._set_hyper_parameter(environment)

        need_to_save_initial_weight = False
        resume_weight_path = self._get_resume_weight_path()
        if resume_weight_path is not None:
            environment.resume_model_weight(resume_weight_path, dataset)
        else:
            initial_weight = self._load_fixed_initial_weight()
            if initial_weight is not None:
                environment.load_model_weight(str(initial_weight), dataset)
            else:
                need_to_save_initial_weight = True

        task = environment.get_train_task()
        if need_to_save_initial_weight:
            self._add_initial_weight_saving_hook(task)

        output_model = environment.get_new_model_entity(dataset)
        score_report_callback = self._prepare_score_report_callback(task)
        task.train(dataset=dataset, output_model=output_model, train_parameters=score_report_callback)
        self._finalize_trial(task)

    def _prepare_hyper_parameter(self):
        return create(self._model_template.hyper_parameters.data)

    def _prepare_dataset_adapter(self):
        dataset_adapter = get_dataset_adapter(
            self._task.task_type,
            self._model_template.hyper_parameters.parameter_overrides["algo_backend"]["train_type"]["default_value"],
            train_data_roots=self._data_roots["train_subset"]["data_root"],
            val_data_roots=self._data_roots["val_subset"]["data_root"] if "val_subset" in self._data_roots else None,
            unlabeled_data_roots=self._data_roots["unlabeled_subset"]["data_root"]
            if "unlabeled_subset" in self._data_roots
            else None,
        )

        return dataset_adapter

    def _set_hyper_parameter(self, environment: TaskEnvironmentManager):
        environment.set_hyper_parameter_using_str_key(self._hp_config["configuration"])
        environment.set_epoch(self._epoch)

    def _prepare_environment(self, hyper_parameters, label_schema):
        enviroment = TaskEnvironment(
            model=None,
            hyper_parameters=hyper_parameters,
            label_schema=label_schema,
            model_template=self._model_template,
        )

        return TaskEnvironmentManager(enviroment)

    def _get_resume_weight_path(self):
        trial_work_dir = self._get_weight_dir_path()
        if not trial_work_dir.exists():
            return None
        return self._task.get_latest_weight(trial_work_dir)

    def _load_fixed_initial_weight(self):
        initial_weight_path = self._get_initial_weight_path()
        if initial_weight_path.exists():
            return initial_weight_path
        return None

    def _add_initial_weight_saving_hook(self, task):
        initial_weight_path = self._get_initial_weight_path()
        task.update_override_configurations(
            {
                "custom_hooks": [
                    dict(
                        type="SaveInitialWeightHook",
                        save_path=initial_weight_path.parent,
                        file_name=initial_weight_path.name,
                    )
                ]
            }
        )

    def _prepare_score_report_callback(self, task) -> TrainParameters:
        return TrainParameters(False, HpoCallback(self._report_func, self._metric, self._epoch, task))

    def _get_initial_weight_path(self) -> Path:
        return self._hpo_workdir / self._initial_weight_name

    def _finalize_trial(self, task):
        weight_dir_path = self._get_weight_dir_path()
        weight_dir_path.mkdir(parents=True, exist_ok=True)
        self._task.copy_weight(task.project_path, weight_dir_path)
        self._report_func(0, 0, done=True)

    def _get_weight_dir_path(self) -> Path:
        return self._hpo_workdir / "weight" / self._hp_config["id"]


def run_trial(
    hp_config: Dict[str, Any],
    report_func: Callable,
    model_template,
    data_roots: Dict[str, str],
    task_type: TaskType,
    hpo_workdir: Union[str, Path],
    initial_weight_name: str,
    metric: str,
):
    """Function to train a model given hyper parameters.

    Args:
        hp_config (Dict[str, Any]): hyper parameter to use on training
        report_func (Callable): function to report score
        model_template: model template
        data_roots (Dict[str, str]): dataset path of each dataset type
        task_type (TaskType): OTX task type
        hpo_workdir (Union[str, Path]): work directory for HPO
        initial_weight_name (str): initial model weight name for each trials to load
        metric (str): metric name
    """
    # pylint: disable=too-many-arguments
    trainer = Trainer(
        hp_config, report_func, model_template, data_roots, task_type, hpo_workdir, initial_weight_name, metric
    )
    trainer.run()


class HpoCallback(UpdateProgressCallback):
    """Callback class to report score to HPO.

    Args:
        report_func (Callable): function to report score
        metric (str): metric name
        max_epoch (int): max_epoch
        task: OTX train task
    """

    def __init__(self, report_func: Callable, metric: str, max_epoch: int, task):
        if max_epoch <= 0:
            raise ValueError(f"max_epoch should be bigger than 0. Current value is {max_epoch}.")

        super().__init__()
        self._report_func = report_func
        self.metric = metric
        self._max_epoch = max_epoch
        self._task = task

    def __call__(self, progress: Union[int, float], score: Optional[float] = None):
        """When callback is called, report a score to HPO algorithm."""
        if score is not None:
            epoch = round(self._max_epoch * progress / 100)
            logger.debug(f"In hpo callback : {score} / {progress} / {epoch}")
            if self._report_func(score=score, progress=epoch) == TrialStatus.STOP:
                self._task.cancel_training()

    def __deepcopy__(self, memo):
        """Prevent repot_func from deepcopied."""
        args = [self.metric, self._max_epoch, self._task]
        copied_args = deepcopy(args, memo)
        return self.__class__(self._report_func, *copied_args)


class HpoDataset:
    """Wrapper class for DatasetEntity of dataset. It's used to make subset during HPO.

    Args:
        fullset: full dataset
        config (Optional[Dict[str, Any]], optional): hyper parameter trial config
        indices (Optional[List[int]]): dataset index. Defaults to None.
    """

    def __init__(self, fullset, config: Optional[Dict[str, Any]] = None, indices: Optional[List[int]] = None):
        self.fullset = fullset
        self.indices = indices
        if config is not None:
            subset_ratio = config["train_environment"]["subset_ratio"]
            self.subset_ratio = 1 if subset_ratio is None else subset_ratio

    def __len__(self) -> int:
        """Get length of subset."""
        if self.indices is None:
            return len(self.fullset)
        return len(self.indices)

    def __getitem__(self, indx) -> dict:
        """Get dataset at index."""
        if self.indices is None:
            return self.fullset[indx]
        return self.fullset[self.indices[indx]]

    def __getattr__(self, name):
        """When trying to get other attributes, not dataset, get values from fullset."""
        if name == "__setstate__":
            raise AttributeError(name)
        return getattr(self.fullset, name)

    def get_subset(self, subset: Subset):
        """Get subset according to subset_ratio if training dataset is requested.

        Args:
            subset (Subset): which subset to get

        Returns:
            HpoDataset: subset wrapped by HpoDataset
        """
        dataset = self.fullset.get_subset(subset)
        if subset != Subset.TRAINING or self.subset_ratio > 0.99:
            return dataset

        indices = torch.randperm(len(dataset), generator=torch.Generator().manual_seed(42))
        indices = indices.tolist()  # type: ignore
        indices = indices[: int(len(dataset) * self.subset_ratio)]

        return HpoDataset(dataset, config=None, indices=indices)



# Copyright (C) 2021 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

"""
The file contains fixtures that may be used in algo backend's
reallife training tests.

Note that the fixtures otx_templates_root_dir_fx and otx_test_domain_fx
MUST be overriden in algo backend's conftest.py file.
"""

# pylint: disable=redefined-outer-name

import glob
import os
import os.path as osp
from copy import deepcopy
from pprint import pformat
from typing import Callable, Dict, Optional

import pytest
import yaml

from otx.api.entities.model_template import parse_model_template

from .e2e_test_system import DataCollector
from .logging import get_logger, set_log_level
from .training_tests_common import REALLIFE_USECASE_CONSTANT, ROOT_PATH_KEY

logger = get_logger()

#########################################################################################
# Fixtures that should be overriden in algo backends


@pytest.fixture(scope="session")
def otx_templates_root_dir_fx():
    """
    The fixture returns an absolute path to the folder where (in the subfolders)
    the reallife training tests will look OTX template files (the files 'template.yaml').

    The fixture MUST be overriden in algo backend's conftest.py file.
    """
    raise NotImplementedError("The fixture otx_templates_root_dir_fx should be overriden in algo backend")


@pytest.fixture
def otx_reference_root_dir_fx():
    """
    The fixture returns an absolute path to the folder where reference files
    for OTX models are stored.
    """
    raise NotImplementedError("The fixture otx_reference_root_dir_fx should be overriden in algo backend")


@pytest.fixture
def otx_current_reference_dir_fx(otx_reference_root_dir_fx, current_test_parameters_fx):
    """
    The fixture returns an absolute path to the folder where reference files
    for the current model are stored.
    """
    if otx_reference_root_dir_fx is None:
        return None
    path = os.path.join(otx_reference_root_dir_fx, current_test_parameters_fx["model_name"])
    if not os.path.isdir(path):
        return None
    return path


@pytest.fixture
def otx_test_domain_fx():
    """
    The fixture returns a string that will be used as the 'subject' field in the
    e2e test system dashboard.
    At the moment it is supposed that the fixture should return something like
    'custom-object-detection'.

    The fixture MUST be overriden in algo backend's conftest.py file.
    """
    raise NotImplementedError("The fixture otx_test_domain_fx should be overriden in algo backend")


@pytest.fixture
def otx_test_scenario_fx():
    """
    The fixture returns a string that will be used as the 'scenario' field in the
    e2e test system dashboard.
    At the moment it is supposed that the fixture should return something like
    'api' or 'integration' or 'reallife'.

    The fixture may be overriden in algo backend's conftest.py file.
    """
    return "api"


#
#########################################################################################


@pytest.fixture
def dataset_definitions_fx(request):
    """
    Return dataset definitions read from a YAML file passed as the parameter --dataset-definitions.

    Note that the dataset definitions should store the following structure:
    {
        <dataset_name1>: { ...<some elements describing dataset1>... },
        <dataset_name2>: { ...<some elements describing dataset2>... },
        ...
    }
    The elements describing datasets could have arbitrary structure, the
    structure is defined by the functions parsing dataset in the algo backends.

    An example for mmdetection algo backend:
    {
        <dataset_name>: {
            'annotations_train': <annotation_file_path1>
            'images_train_dir': <images_folder_path1>
            'annotations_val': <annotation_file_path2>
            'images_val_dir': <images_folder_path2>
            'annotations_test': <annotation_file_path3>
            'images_test_dir':  <images_folder_path3>
        }
    }

    Also one more key with value ROOT_PATH_KEY is added -- it is the path to
    the folder where the dataset definitions file is placed, this path will be
    used to resolve relative paths in the dataset structures.
    """
    path = request.config.getoption("--dataset-definitions")
    if path is None:
        logger.warning(
            f"The command line parameter '--dataset-definitions' is not set"
            f"whereas it is required for the test {request.node.originalname or request.node.name}"
            f" -- ALL THE TESTS THAT REQUIRE THIS PARAMETER ARE SKIPPED"
        )
        return None
    with open(path, encoding="utf-8") as f:
        data = yaml.safe_load(f)
    data[ROOT_PATH_KEY] = osp.dirname(path)
    return data


@pytest.fixture(scope="session")
def template_paths_fx(otx_templates_root_dir_fx):
    """
    Return mapping model names to template paths, received from globbing the
    folder pointed by the fixture otx_templates_root_dir_fx.
    Note that the function searches files with name `template.yaml`, and for each such file
    the model name is the name of the parent folder of the file.
    """
    root = otx_templates_root_dir_fx
    assert osp.isabs(root), f"Error: otx_templates_root_dir_fx is not an absolute path: {root}"
    template_glob = glob.glob(f"{root}/**/template*.yaml", recursive=True)
    data = {}
    for cur_path in template_glob:
        assert osp.isabs(cur_path), f"Error: not absolute path {cur_path}"
        name = parse_model_template(cur_path).model_template_id
        if name in data:
            raise RuntimeError(f"Duplication of names in {root} folder: {data[name]} and {cur_path}")
        assert name != ROOT_PATH_KEY, f"Wrong model name {name}"
        data[name] = cur_path
    data[ROOT_PATH_KEY] = ""
    return data


@pytest.fixture
def expected_metrics_all_tests_fx(request):
    # pylint: disable=line-too-long
    """
    Return expected metrics for reallife tests read from a YAML file passed as the parameter --expected-metrics-file.
    Note that the structure of expected metrics should be a dict that maps tests to the expected metric numbers.
    The keys of the dict are the parameters' part of the test id-s -- see the function
    OTXTestHelper._generate_test_id, also see the fixture current_test_parameters_string_fx below.

    The value for each key is a structure that stores a requirement on some metric.
    The requirement can be either a target value (probably, with max size of quality drop)
    or the reference to another stage of the same model (also probably with max size of quality drop).
    See details in the description of the fixture cur_test_expected_metrics_callback_fx below.
    E.g.
    ```
    'ACTION-training_evaluation,model-gen3_mobilenetV2_ATSS,dataset-bbcd,num_iters-KEEP_CONFIG_FIELD_VALUE,batch-KEEP_CONFIG_FIELD_VALUE,usecase-reallife':
        'metrics.accuracy.f-measure':
            'target_value': 0.81
            'max_diff': 0.005
    'ACTION-export_evaluation,model-gen3_mobilenetV2_ATSS,dataset-bbcd,num_iters-KEEP_CONFIG_FIELD_VALUE,batch-KEEP_CONFIG_FIELD_VALUE,usecase-reallife':
        'metrics.accuracy.f-measure':
            'base': 'training_evaluation.metrics.accuracy.f-measure'
            'max_diff': 0.01
    ```
    """
    path = request.config.getoption("--expected-metrics-file")
    if path is None:
        logger.warning(
            "The command line parameter '--expected-metrics-file' is not set"
            "whereas it is required to compare with target metrics"
            " -- ALL THE COMPARISON WITH TARGET METRICS IN TESTS WILL BE FAILED"
        )
        return None
    with open(path, encoding="utf-8") as f:
        expected_metrics_all_tests = yaml.safe_load(f)
    assert isinstance(expected_metrics_all_tests, dict), f"Wrong metrics file {path}: {expected_metrics_all_tests}"
    return expected_metrics_all_tests


@pytest.fixture(scope="session", autouse=True)
def force_logging_session_fx(request):
    """
    This fixture force setting log level for test suite.
    It may be required in the case when one of the packages
    sets global log level to logging.ERROR.
    This fixture has session scope.
    """
    level = request.config.getoption("--force-log-level")
    recursive_level = request.config.getoption("--force-log-level-recursive")
    if recursive_level is not None:
        set_log_level(recursive_level, recursive=True)
    if level is not None:
        set_log_level(level)


@pytest.fixture
def force_logging_fx(request):
    """
    This fixture force setting log level for test suite.
    It may be required in the case when one of the packages
    sets global log level to logging.ERROR.
    Note that using --force-log-level-recursive option
    it is possible to set log level for all parents of the test
    suite logger.
    This fixture has function scope -- it may be required if some
    of packages changes log level of some loggers during work of test.
    """
    level = request.config.getoption("--force-log-level")
    recursive_level = request.config.getoption("--force-log-level-recursive")
    if recursive_level is not None:
        set_log_level(recursive_level, recursive=True)
    if level is not None:
        set_log_level(level)


@pytest.fixture
def current_test_parameters_fx(request, force_logging_fx):
    # pylint: disable=unused-argument
    """
    This fixture returns the test parameter `test_parameters` of the current test.
    """
    cur_test_params = deepcopy(request.node.callspec.params)
    assert "test_parameters" in cur_test_params, (
        f"The test {request.node.name} should be parametrized " f"by parameter 'test_parameters'"
    )
    return cur_test_params["test_parameters"]


@pytest.fixture
def current_test_parameters_string_fx(request, force_logging_fx):
    # pylint: disable=unused-argument
    """
    This fixture returns the part of the test id between square brackets
    (i.e. the part of id that corresponds to the test parameters)
    """
    node_name = request.node.name
    assert "[" in node_name, f"Wrong format of node name {node_name}"
    assert node_name.endswith("]"), f"Wrong format of node name {node_name}"
    index = node_name.find("[")
    return node_name[index + 1 : -1]


# TODO(lbeynens): replace 'callback' with 'factory'
@pytest.fixture
def cur_test_expected_metrics_callback_fx(
    expected_metrics_all_tests_fx,
    current_test_parameters_string_fx,
    current_test_parameters_fx,
) -> Optional[Callable[[], Dict]]:
    """
    This fixture returns
    * either a callback -- a function without parameters that returns
      expected metrics for the current test,
    * or None if the test validation should be skipped.

    The expected metrics for a test is a dict with the structure that stores the
    requirements on metrics on the current test. In this dict
    * each key is a dot-separated metric "address" in the structure received as the result of the test
    * each value is a structure describing a requirement for this metric
    e.g.
    ```
    {
      'metrics.accuracy.f-measure': {
              'target_value': 0.81,
              'max_diff': 0.005
          }
    }
    ```

    Note that the fixture returns a callback instead of returning the expected metrics structure
    themselves, to avoid attempts to read expected metrics for the stages that do not make validation
    at all -- now the callback is called if and only if validation is made for the stage.
    (E.g. the stage 'export' does not make validation, but the stage 'export_evaluation' does.)

    Also note that if the callback is called, but the expected metrics for the current test
    are not found in the structure with expected metrics for all tests, then the callback
    raises exception ValueError to fail the test.

    And also note that each requirement for each metric is a dict with the following structure:
    * The dict points a target value of the metric.
      The target_value may be pointed
      ** either by key 'target_value' (in this case the value is float),
      ** or by the key 'base', in this case the value is a dot-separated address to another value in the
         storage of previous stages' results, e.g.
             'base': 'training_evaluation.metrics.accuracy.f-measure'

    * The dict points a range of acceptable values for the metric.
      The range for the metric values may be pointed
      ** either by key 'max_diff' (with float value),
         in this case the acceptable range will be
         [target_value - max_diff, target_value + max_diff]
         (inclusively).

      ** or the range may be pointed by keys 'max_diff_if_less_threshold' and/or 'max_diff_if_greater_threshold'
         (with float values), in this case the acceptable range is
         `[target_value - max_diff_if_less_threshold, target_value + max_diff_if_greater_threshold]`
         (also inclusively).
         This allows to point non-symmetric ranges w.r.t. the target_value.
         One of 'max_diff_if_less_threshold' or 'max_diff_if_greater_threshold' may be absent, in this case
         it is set to `+infinity`, so the range will be half-bounded.
         E.g. if `max_diff_if_greater_threshold` is absent, the range will be
         [target_value - max_diff_if_less_threshold, +infinity]
    """
    if REALLIFE_USECASE_CONSTANT != current_test_parameters_fx["usecase"]:
        return None

    # make a copy to avoid later changes in the structs
    expected_metrics_all_tests = deepcopy(expected_metrics_all_tests_fx)
    current_test_parameters_string = deepcopy(current_test_parameters_string_fx)

    def _get_expected_metrics_callback():
        if expected_metrics_all_tests is None:
            raise ValueError(
                f"The dict with expected metrics cannot be read, although it is required "
                f"for validation in the test '{current_test_parameters_string}'"
            )
        if current_test_parameters_string not in expected_metrics_all_tests:
            raise ValueError(
                f"The parameters id string {current_test_parameters_string} is not inside "
                f"the dict with expected metrics -- cannot make validation, so test is failed"
            )
        expected_metrics = expected_metrics_all_tests[current_test_parameters_string]
        if not isinstance(expected_metrics, dict):
            raise ValueError(
                f"The expected metric for parameters id string {current_test_parameters_string} "
                f"should be a dict, whereas it is: {pformat(expected_metrics)}"
            )
        return expected_metrics

    return _get_expected_metrics_callback


@pytest.fixture
def data_collector_fx(request, otx_test_scenario_fx, otx_test_domain_fx) -> DataCollector:
    """
    The fixture returns the DataCollector instance that may be used to pass
    the values (metrics, intermediate results, etc) to the e2e test system dashboard.
    Please, see the interface of DataCollector class in the function
    e2e_test_system._create_class_DataCollector
    (the function creates a stub class with the proper interface if e2e test system is not installed).

    Note that the fixture contains both setup and teardown parts using yield from fixture.
    Each test uses its own instance of DataCollector class, so each test will create its own row in the
    dashboard of e2e test system.
    """
    setup = deepcopy(request.node.callspec.params)
    setup["environment_name"] = os.environ.get("TT_ENVIRONMENT_NAME", "no-env")
    setup["test_type"] = os.environ.get("TT_TEST_TYPE", "no-test-type")  # TODO: get from e2e test type
    setup["scenario"] = otx_test_scenario_fx
    setup["test"] = request.node.name
    setup["subject"] = otx_test_domain_fx
    setup["project"] = "otx"
    if "test_parameters" in setup:
        assert isinstance(setup["test_parameters"], dict)
        if "dataset_name" not in setup:
            setup["dataset_name"] = setup["test_parameters"].get("dataset_name")
        if "model_name" not in setup:
            setup["model_name"] = setup["test_parameters"].get("model_name")
        if "test_stage" not in setup:
            setup["test_stage"] = setup["test_parameters"].get("test_stage")
        if "usecase" not in setup:
            setup["usecase"] = setup["test_parameters"].get("usecase")
    logger.info(f"creating DataCollector: setup=\n{pformat(setup, width=140)}")
    data_collector = DataCollector(name="TestOTXIntegration", setup=setup)
    with data_collector:
        logger.info("data_collector is created")
        yield data_collector
    logger.info("data_collector is released")



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import functools

from mmseg.models import SEGMENTORS
from mmseg.utils import get_root_logger

from otx.mpa.modules.utils.task_adapt import map_class_names

from .mix_loss_mixin import MixLossMixin
from .otx_encoder_decoder import OTXEncoderDecoder
from .pixel_weights_mixin import PixelWeightsMixin


@SEGMENTORS.register_module()
class ClassIncrEncoderDecoder(MixLossMixin, PixelWeightsMixin, OTXEncoderDecoder):
    """ """

    def __init__(self, *args, task_adapt=None, **kwargs):
        super().__init__(*args, **kwargs)

        # Hook for class-sensitive weight loading
        assert task_adapt is not None, "When using task_adapt, task_adapt must be set."

        self._register_load_state_dict_pre_hook(
            functools.partial(
                self.load_state_dict_pre_hook,
                self,  # model
                task_adapt["dst_classes"],  # model_classes
                task_adapt["src_classes"],  # chkpt_classes
            )
        )

    @staticmethod
    def load_state_dict_pre_hook(model, model_classes, chkpt_classes, chkpt_dict, prefix, *args, **kwargs):
        """Modify input state_dict according to class name matching before weight loading"""
        logger = get_root_logger("INFO")
        logger.info(f"----------------- ClassIncrEncoderDecoder.load_state_dict_pre_hook() called w/ prefix: {prefix}")

        # Dst to src mapping index
        model_classes = list(model_classes)
        chkpt_classes = list(chkpt_classes)
        model2chkpt = map_class_names(model_classes, chkpt_classes)
        logger.info(f"{chkpt_classes} -> {model_classes} ({model2chkpt})")

        model_dict = model.state_dict()
        param_names = [
            "decode_head.conv_seg.weight",
            "decode_head.conv_seg.bias",
        ]
        for model_name in param_names:
            chkpt_name = prefix + model_name
            if model_name not in model_dict or chkpt_name not in chkpt_dict:
                logger.info(f"Skipping weight copy: {chkpt_name}")
                continue

            # Mix weights
            model_param = model_dict[model_name].clone()
            chkpt_param = chkpt_dict[chkpt_name]
            for m, c in enumerate(model2chkpt):
                if c >= 0:
                    model_param[m].copy_(chkpt_param[c])

            # Replace checkpoint weight by mixed weights
            chkpt_dict[chkpt_name] = model_param



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import numpy as np

from otx.algorithms.classification.adapters.openvino.model_wrappers import (
    OTXClassification,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.api.parameters_validation.validation_helper import (
    check_value_error_exception_raised,
)


class MockClassification(OTXClassification):
    def __init__(self):
        pass


class TestOTXClassificationParamsValidation:
    @e2e_pytest_unit
    def test_ote_classification_postprocess_params_validation(self):
        """
        <b>Description:</b>
        Check OTXClassification object "postprocess" method input parameters validation

        <b>Input data:</b>
        OTXClassification object. "postprocess" method unexpected type parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "postprocess" method
        """
        classification = MockClassification()
        correct_values_dict = {
            "outputs": {"output_1": np.random.rand(2, 2)},
            "metadata": {"metadata_1": "some_data"},
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "outputs" parameter
            ("outputs", unexpected_int),
            # Unexpected integer is specified as "outputs" dictionary key
            ("outputs", {unexpected_int: np.random.rand(2, 2)}),
            # Unexpected integer is specified as "outputs" dictionary value
            ("outputs", {"output_1": unexpected_int}),
            # Unexpected integer is specified as "metadata" parameter
            ("metadata", unexpected_int),
            # Unexpected integer is specified as "metadata" dictionary key
            ("metadata", {unexpected_int: "some_data"}),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=classification.postprocess,
        )

    @e2e_pytest_unit
    def test_ote_classification_postprocess_aux_outputs_params_validation(self):
        """
        <b>Description:</b>
        Check OTXClassification object "postprocess_aux_outputs" method input parameters validation

        <b>Input data:</b>
        OTXClassification object. "postprocess_aux_outputs" method unexpected type parameters

        <b>Expected results:</b>
        Test passes if ValueError exception is raised when unexpected type object is specified as
        input parameter for "postprocess_aux_outputs" method
        """
        classification = MockClassification()
        correct_values_dict = {
            "outputs": {"output_1": np.random.rand(2, 2)},
            "metadata": {"metadata_1": "some_data"},
        }
        unexpected_int = 1
        unexpected_values = [
            # Unexpected integer is specified as "outputs" parameter
            ("outputs", unexpected_int),
            # Unexpected integer is specified as "outputs" dictionary key
            ("outputs", {unexpected_int: np.random.rand(2, 2)}),
            # Unexpected integer is specified as "outputs" dictionary value
            ("outputs", {"output_1": unexpected_int}),
            # Unexpected integer is specified as "metadata" parameter
            ("metadata", unexpected_int),
            # Unexpected integer is specified as "metadata" dictionary key
            ("metadata", {unexpected_int: "some_data"}),
        ]
        check_value_error_exception_raised(
            correct_parameters=correct_values_dict,
            unexpected_values=unexpected_values,
            class_or_function=classification.postprocess_aux_outputs,
        )



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import copy

import pytest

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.elements import metadata_keys
from otx.api.configuration.enums import AutoHPOState
from otx.api.configuration.enums.config_element_type import ConfigElementType
from otx.api.entities.id import ID
from tests.unit.api.configuration.dummy_config import DatasetManagerConfig
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestConfigurableParameters:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_configurable_parameters(self):
        """
        <b>Description:</b>
        Check "ConfigurableParameters" class object initialization

        <b>Input data:</b>
        "ConfigurableParameters" class object with specified initialization parameters

        <b>Expected results:</b>
        Test passes if attributes of initialized "ConfigurableParameters" class object are equal to expected
        """

        def check_configurable_parameters_attributes(
            configurable_parameters: ConfigurableParameters,
            expected_header: str,
            expected_description: str,
            expected_id: ID,
            expected_visible_in_ui: bool,
        ):
            assert configurable_parameters.header == expected_header
            assert configurable_parameters.description == expected_description
            assert configurable_parameters.type == ConfigElementType.CONFIGURABLE_PARAMETERS
            assert configurable_parameters.groups == []
            assert configurable_parameters.id == expected_id
            assert configurable_parameters.visible_in_ui == expected_visible_in_ui

        header = "Test Header"
        # Checking "ConfigurableParameters" initialized with default optional parameters
        check_configurable_parameters_attributes(
            configurable_parameters=ConfigurableParameters(header=header),
            expected_header=header,
            expected_description="Default parameter group description",
            expected_id=ID(""),
            expected_visible_in_ui=True,
        )
        # Checking "ConfigurableParameters" initialized with specified optional parameters
        description = "Test Description"
        config_id = ID("Test ID")
        visible_in_ui = False
        check_configurable_parameters_attributes(
            configurable_parameters=ConfigurableParameters(
                header=header,
                description=description,
                id=config_id,
                visible_in_ui=visible_in_ui,
            ),
            expected_header=header,
            expected_description=description,
            expected_id=config_id,
            expected_visible_in_ui=visible_in_ui,
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_set_metadata(self):
        """
        <b>Description:</b>
        Check "ConfigurableParameters" class parameter metadata setting

        <b>Input data:</b>
        Dummy configuration -- DatasetManagerConfig

        <b>Expected results:</b>
        Test passes if:
            1. Metadata for a given parameter inside the ConfigurableParameters can be
               set successfully,
            2. Attempting to set metadata for a non-existing parameter results in
               failure
            3. Attempting to set metadata for a non-existing metadata key results in
               failure
            4. Attempting to set metadata to a value of a type that does not match the
               original metadata item type results in failure
            5. Resetting the metadata back to its original value can be done
               successfully
        """
        # Arrange
        config = DatasetManagerConfig(
            description="Configurable parameters for the DatasetManager -- TEST ONLY",
            header="Dataset Manager configuration -- TEST ONLY",
        )
        test_parameter_name = "dummy_float_selectable"
        metadata_key = metadata_keys.AUTO_HPO_STATE
        old_value = config.get_metadata(test_parameter_name)[metadata_key]
        new_value = AutoHPOState.OPTIMIZED

        # Act
        success = config.set_metadata_value(
            parameter_name=test_parameter_name,
            metadata_key=metadata_key,
            value=new_value,
        )
        no_success_invalid_param = config.set_metadata_value(
            parameter_name=test_parameter_name + "_invalid",
            metadata_key=metadata_key,
            value=new_value,
        )
        no_success_invalid_key = config.set_metadata_value(
            parameter_name=test_parameter_name,
            metadata_key=metadata_key + "_invalid",
            value=new_value,
        )
        no_success_invalid_value_type = config.set_metadata_value(
            parameter_name=test_parameter_name,
            metadata_key=metadata_key,
            value=str(new_value),
        )
        config_copy = copy.deepcopy(config)
        success_revert = config_copy.set_metadata_value(
            parameter_name=test_parameter_name,
            metadata_key=metadata_key,
            value=old_value,
        )

        # Assert
        assert old_value != new_value
        assert all([success, success_revert])
        assert not any(
            [
                no_success_invalid_key,
                no_success_invalid_param,
                no_success_invalid_value_type,
            ]
        )
        assert config.get_metadata(test_parameter_name)[metadata_key] == new_value
        assert config_copy.get_metadata(test_parameter_name)[metadata_key] == old_value

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_update_auto_hpo_state(self):
        """
        <b>Description:</b>
        Check that updating the auto_hpo_state for all parameters within a
        "ConfigurableParameters" class parameter works as expected

        <b>Input data:</b>
        Dummy configuration -- DatasetManagerConfig

        <b>Expected results:</b>
        Test passes if:
            1. The `auto_hpo_state` metadata field is updated to `optimized` upon
                calling config.update_auto_hpo_states() if the value for a
                configurable parameter matches that of its `auto_hpo_value`
                metadata field
            2. The `auto_hpo_state` is updated to `overridden` upon calling
                config.update_hpo_state() if the value for a configurable parameter
                does not match that of its `auto_hpo_value` metadata field
        """
        # Arrange
        config = DatasetManagerConfig(
            description="Configurable parameters for the DatasetManager -- TEST ONLY",
            header="Dataset Manager configuration -- TEST ONLY",
        )
        test_parameter_1 = "dummy_float_selectable"
        test_parameter_2 = "train_proportion"
        auto_hpo_result_float = 4.0
        auto_hpo_result_train_prop = 0.9

        config.dummy_float_selectable = auto_hpo_result_float
        success_1 = config.set_metadata_value(
            parameter_name=test_parameter_1,
            metadata_key=metadata_keys.AUTO_HPO_VALUE,
            value=auto_hpo_result_float,
        )
        config.subset_parameters.train_proportion = auto_hpo_result_train_prop
        success_2 = config.subset_parameters.set_metadata_value(
            parameter_name=test_parameter_2,
            metadata_key=metadata_keys.AUTO_HPO_VALUE,
            value=auto_hpo_result_train_prop,
        )

        # Act
        config.update_auto_hpo_states()
        auto_hpo_state_1 = config.get_metadata(test_parameter_1)[metadata_keys.AUTO_HPO_STATE]
        auto_hpo_state_2 = config.subset_parameters.get_metadata(test_parameter_2)[metadata_keys.AUTO_HPO_STATE]

        # Simulate override
        config.dummy_float_selectable = auto_hpo_result_float - 1
        config.subset_parameters.train_proportion = auto_hpo_result_train_prop - 0.001

        config.update_auto_hpo_states()
        auto_hpo_state_override_1 = config.get_metadata(test_parameter_1)[metadata_keys.AUTO_HPO_STATE]
        auto_hpo_state_override_2 = config.subset_parameters.get_metadata(test_parameter_2)[
            metadata_keys.AUTO_HPO_STATE
        ]

        # Assert
        assert all([success_1, success_2])
        assert auto_hpo_state_1 == AutoHPOState.OPTIMIZED
        assert auto_hpo_state_2 == AutoHPOState.OPTIMIZED
        assert auto_hpo_state_override_1 == AutoHPOState.OVERRIDDEN
        assert auto_hpo_state_override_2 == AutoHPOState.OVERRIDDEN



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from io import BytesIO

import pytest
import torch

from otx.algorithms.common.adapters.nncf.compression import NNCFMetaState
from otx.algorithms.common.adapters.nncf.utils.utils import (
    _is_nncf_enabled,
    check_nncf_is_enabled,
    is_accuracy_aware_training_set,
    is_in_nncf_tracing,
    is_nncf_enabled,
    load_checkpoint,
    nncf_trace,
    no_nncf_trace,
    nullcontext,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.common.adapters.mmcv.nncf.test_helpers import create_model


@e2e_pytest_unit
def test_is_nncf_enabled():
    assert _is_nncf_enabled == is_nncf_enabled()


@e2e_pytest_unit
def test_nncf_is_enabled():
    if is_nncf_enabled():
        check_nncf_is_enabled()
    else:
        with pytest.raises(RuntimeError):
            check_nncf_is_enabled()


@e2e_pytest_unit
def test_load_checkpoint():
    mock_model = create_model()
    state_dict = mock_model.state_dict()
    buffer = BytesIO()
    torch.save(state_dict, buffer)
    buffer.seek(0)

    load_checkpoint(mock_model, buffer, strict=False)
    assert state_dict.keys() == mock_model.state_dict().keys()
    for k in state_dict.keys():
        assert torch.equal(state_dict[k], mock_model.state_dict()[k])

    buffer = BytesIO()
    torch.save(
        {
            "state_dict": {"dummy": "dummy"},
            "meta": {"nncf_meta": NNCFMetaState(state_to_build=state_dict, compression_ctrl={"dummy": "dummy"})},
        },
        buffer,
    )
    buffer.seek(0)
    load_checkpoint(mock_model, buffer, strict=False)
    assert state_dict.keys() == mock_model.state_dict().keys()

    state_dict["dummy"] = torch.tensor(1)
    buffer = BytesIO()
    torch.save(state_dict, buffer)
    buffer.seek(0)
    with pytest.raises(RuntimeError):
        load_checkpoint(mock_model, buffer, strict=True)


@e2e_pytest_unit
def test_nullcontext():
    with nullcontext():
        pass


@e2e_pytest_unit
def test_no_nncf_trace():
    with no_nncf_trace():
        pass


@e2e_pytest_unit
def test_nncf_trace():
    with nncf_trace():
        pass


@e2e_pytest_unit
def test_is_in_nncf_tracing():
    assert is_in_nncf_tracing() is False


@e2e_pytest_unit
def test_is_accuracy_aware_training_set():
    assert is_accuracy_aware_training_set({"accuracy_aware_training": True}) is True
    assert is_accuracy_aware_training_set({}) is False
    assert is_accuracy_aware_training_set({"accuracy_aware_training": {}}) is True



# Copyright (C) 2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import torch
from mmdet.core import images_to_levels, multi_apply
from mmdet.models.builder import HEADS
from mmdet.models.dense_heads.base_dense_head import BaseDenseHead


@HEADS.register_module()
class CrossDatasetDetectorHead(BaseDenseHead):
    def get_atss_targets(
        self,
        anchor_list,
        valid_flag_list,
        gt_bboxes_list,
        img_metas,
        gt_bboxes_ignore_list=None,
        gt_labels_list=None,
        label_channels=1,
        unmap_outputs=True,
    ):
        """Get targets for Detection head.

        This method is almost the same as `AnchorHead.get_targets()`. Besides
        returning the targets as the parent method does, it also returns the
        anchors as the first element of the returned tuple.
        However, if the detector's head loss uses CrossSigmoidFocalLoss,
        the labels_weights_list consists of (binarized label schema * weights) of batch images
        """
        num_imgs = len(img_metas)
        assert len(anchor_list) == len(valid_flag_list) == num_imgs

        # anchor number of multi levels
        num_level_anchors = [anchors.size(0) for anchors in anchor_list[0]]
        num_level_anchors_list = [num_level_anchors] * num_imgs

        # concat all level anchors and flags to a single tensor
        for i in range(num_imgs):
            assert len(anchor_list[i]) == len(valid_flag_list[i])
            anchor_list[i] = torch.cat(anchor_list[i])
            valid_flag_list[i] = torch.cat(valid_flag_list[i])

        # compute targets for each image
        if gt_bboxes_ignore_list is None:
            gt_bboxes_ignore_list = [None for _ in range(num_imgs)]
        if gt_labels_list is None:
            gt_labels_list = [None for _ in range(num_imgs)]
        (
            all_anchors,
            all_labels,
            all_label_weights,
            all_bbox_targets,
            all_bbox_weights,
            pos_inds_list,
            neg_inds_list,
        ) = multi_apply(
            self._get_target_single,
            anchor_list,
            valid_flag_list,
            num_level_anchors_list,
            gt_bboxes_list,
            gt_bboxes_ignore_list,
            gt_labels_list,
            img_metas,
            label_channels=label_channels,
            unmap_outputs=unmap_outputs,
        )
        # no valid anchors
        if any([labels is None for labels in all_labels]):
            return None
        # sampled anchors of all images
        num_total_pos = sum([max(inds.numel(), 1) for inds in pos_inds_list])
        num_total_neg = sum([max(inds.numel(), 1) for inds in neg_inds_list])
        # split targets to a list w.r.t. multiple levels
        anchors_list = images_to_levels(all_anchors, num_level_anchors)
        labels_list = images_to_levels(all_labels, num_level_anchors)
        valid_label_mask = self.get_valid_label_mask(img_metas=img_metas, all_labels=all_labels)
        valid_label_mask = [i.to(anchor_list[0].device) for i in valid_label_mask]
        if len(valid_label_mask) > 0:
            valid_label_mask = images_to_levels(valid_label_mask, num_level_anchors)

        label_weights_list = images_to_levels(all_label_weights, num_level_anchors)
        bbox_targets_list = images_to_levels(all_bbox_targets, num_level_anchors)
        bbox_weights_list = images_to_levels(all_bbox_weights, num_level_anchors)
        return (
            anchors_list,
            labels_list,
            label_weights_list,
            bbox_targets_list,
            bbox_weights_list,
            valid_label_mask,
            num_total_pos,
            num_total_neg,
        )

    def get_fcos_targets(self, points, gt_bboxes_list, gt_labels_list, img_metas):
        """Compute regression, classification and centerss targets for points
        in multiple images.

        Args:
            points (list[Tensor]): Points of each fpn level, each has shape
                (num_points, 2).
            gt_bboxes_list (list[Tensor]): Ground truth bboxes of each image,
                each has shape (num_gt, 4).
            gt_labels_list (list[Tensor]): Ground truth labels of each box,
                each has shape (num_gt,).

        Returns:
            tuple:
                concat_lvl_labels (list[Tensor]): Labels of each level. \
                concat_lvl_bbox_targets (list[Tensor]): BBox targets of each \
                    level.
        """
        assert len(points) == len(self.regress_ranges)
        num_levels = len(points)
        # expand regress ranges to align with points
        expanded_regress_ranges = [
            points[i].new_tensor(self.regress_ranges[i])[None].expand_as(points[i]) for i in range(num_levels)
        ]
        # concat all levels points and regress ranges
        concat_regress_ranges = torch.cat(expanded_regress_ranges, dim=0)
        concat_points = torch.cat(points, dim=0)

        # the number of points per img, per lvl
        num_points = [center.size(0) for center in points]

        # get labels and bbox_targets of each image
        labels_list, bbox_targets_list = multi_apply(
            self._get_target_single,
            gt_bboxes_list,
            gt_labels_list,
            points=concat_points,
            regress_ranges=concat_regress_ranges,
            num_points_per_lvl=num_points,
        )

        # split to per img, per level
        valid_label_mask = self.get_valid_label_mask(img_metas=img_metas, all_labels=labels_list)
        valid_label_mask = [i.to(points[0].device) for i in valid_label_mask]
        labels_list = [labels.split(num_points, 0) for labels in labels_list]
        bbox_targets_list = [bbox_targets.split(num_points, 0) for bbox_targets in bbox_targets_list]

        # concat per level image
        concat_lvl_labels = []
        concat_lvl_bbox_targets = []
        for i in range(num_levels):
            concat_lvl_labels.append(torch.cat([labels[i] for labels in labels_list]))
            bbox_targets = torch.cat([bbox_targets[i] for bbox_targets in bbox_targets_list])
            if self.norm_on_bbox:
                bbox_targets = bbox_targets / self.strides[i]
            concat_lvl_bbox_targets.append(bbox_targets)

        label_weights, bbox_weights = None, None
        return concat_lvl_labels, label_weights, concat_lvl_bbox_targets, bbox_weights, valid_label_mask

    def vfnet_to_atss_targets(self, cls_scores, mlvl_points, gt_bboxes, gt_labels, img_metas, gt_bboxes_ignore=None):
        """A wrapper for computing ATSS targets for points in multiple images.

        Args:
            cls_scores (list[Tensor]): Box iou-aware scores for each scale
                level with shape (N, num_points * num_classes, H, W).
            mlvl_points (list[Tensor]): Points of each fpn level, each has
                shape (num_points, 2).
            gt_bboxes (list[Tensor]): Ground truth bboxes of each image,
                each has shape (num_gt, 4).
            gt_labels (list[Tensor]): Ground truth labels of each box,
                each has shape (num_gt,).
            img_metas (list[dict]): Meta information of each image, e.g.,
                image size, scaling factor, etc.
            gt_bboxes_ignore (None | Tensor): Ground truth bboxes to be
                ignored, shape (num_ignored_gts, 4). Default: None.

        Returns:
            tuple:
                labels_list (list[Tensor]): Labels of each level.
                label_weights (Tensor): Label weights of all levels.
                bbox_targets_list (list[Tensor]): Regression targets of each
                    level, (l, t, r, b).
                bbox_weights (Tensor): Bbox weights of all levels.
        """
        featmap_sizes = [featmap.size()[-2:] for featmap in cls_scores]
        assert len(featmap_sizes) == self.anchor_generator.num_levels

        device = cls_scores[0].device
        anchor_list, valid_flag_list = self.get_anchors(featmap_sizes, img_metas, device=device)
        label_channels = self.cls_out_channels if self.use_sigmoid_cls else 1

        cls_reg_targets = self.get_atss_targets(
            anchor_list,
            valid_flag_list,
            gt_bboxes,
            img_metas,
            gt_bboxes_ignore_list=gt_bboxes_ignore,
            gt_labels_list=gt_labels,
            label_channels=label_channels,
            unmap_outputs=True,
        )
        if cls_reg_targets is None:
            return None

        (
            anchor_list,
            labels_list,
            label_weights_list,
            bbox_targets_list,
            bbox_weights_list,
            valid_label_mask,
            num_total_pos,
            num_total_neg,
        ) = cls_reg_targets

        bbox_targets_list = [bbox_targets.reshape(-1, 4) for bbox_targets in bbox_targets_list]

        num_imgs = len(img_metas)
        # transform bbox_targets (x1, y1, x2, y2) into (l, t, r, b) format
        bbox_targets_list = self.transform_bbox_targets(bbox_targets_list, mlvl_points, num_imgs)

        labels_list = [labels.reshape(-1) for labels in labels_list]
        label_weights_list = [label_weights.reshape(-1) for label_weights in label_weights_list]
        bbox_weights_list = [bbox_weights.reshape(-1) for bbox_weights in bbox_weights_list]
        label_weights = torch.cat(label_weights_list)
        bbox_weights = torch.cat(bbox_weights_list)
        return labels_list, label_weights, bbox_targets_list, bbox_weights, valid_label_mask

    def get_valid_label_mask(self, img_metas, all_labels, use_bg=False):
        num_classes = self.num_classes + 1 if use_bg else self.num_classes
        valid_label_mask = []
        for i, meta in enumerate(img_metas):
            mask = torch.Tensor([1 for _ in range(num_classes)])
            if "ignored_labels" in meta and len(meta["ignored_labels"]) > 0:
                mask[meta["ignored_labels"]] = 0
                if use_bg:
                    mask[self.num_classes] = 0
            mask = mask.repeat(len(all_labels[i]), 1)
            valid_label_mask.append(mask)
        return valid_label_mask



"""Definitions for `substitute_values` and `substitute_values_for_lifecycle` functions within the configuration helper.

These functions can be used to update values or ids in a OTX configuration object, according to a value dictionary or
configuration object
"""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#


from typing import Dict, Optional, Sequence, Union

from omegaconf import DictConfig

from otx.api.configuration.configurable_parameters import ConfigurableParameters
from otx.api.configuration.elements import ParameterGroup, metadata_keys
from otx.api.configuration.enums import ModelLifecycle

from .convert import convert
from .create import input_to_config_dict
from .utils import search_in_config_dict
from .validate import validate


def _should_parameter_be_updated(
    config_element: ParameterGroup,
    parameter_name: str,
    model_lifecycle: Optional[Union[Sequence[ModelLifecycle], ModelLifecycle]],
) -> bool:
    """Check if parameter should be updated.

    Checks whether a parameter `parameter_name` belonging to a ParameterGroup
    `config_element` fullfills the criterion for updating it's value, giving in
    `model_lifecycle`.

    config_element (ParameterGroup): ParameterGroup which the parameter belongs to
    parameter_name (str): Name of the parameter
    model_lifecycle (Optional[Union[Sequence[ModelLifecycle], ModelLifecycle]]):
        Phase or list of phases of the model lifecycle which the parameter should
        affect in order to update it.

    Returns:
        bool: True if the parameter should be updated, False otherwise
    """
    should_update = True
    if model_lifecycle is not None:
        # Skip parameter substitution if model lifecycle does not match
        # provided value
        metadata = config_element.get_metadata(parameter_name)
        parameter_affects_outcome_of = metadata[metadata_keys.AFFECTS_OUTCOME_OF]
        if isinstance(model_lifecycle, list):
            should_update = parameter_affects_outcome_of in model_lifecycle
        else:
            should_update = parameter_affects_outcome_of == model_lifecycle
    return should_update


def _substitute(
    config: ConfigurableParameters,
    value_input: Dict,
    allow_missing_values: bool = False,
    model_lifecycle: Optional[Union[ModelLifecycle, Sequence[ModelLifecycle]]] = None,
):
    """Substitutes values from value_input into the config object.

    The structures of value_input and config have to match in order for the values to be substituted
    correctly. If the argument `model_lifecycle` is provided, only parameters that
    affect the specified phase in the model lifecycle will be substituted.

    Values are substituted in place.

    Args:
        config (ConfigurableParameters): ConfigurableParameter object to substitute values into.
        value_input (Dict): ConfigurableParameters (either in object, dict, yaml or DictConfig representation) to take
            the values to be substituted from.
        allow_missing_values (bool): True to allow missing values in the configuration, i.e. if a value is found in
            `value_input`, but not in `config`, it will silently be ignored. If set to False, an AttributeError will be
            raised. Defaults to False.
        model_lifecycle (Optional[Union[ModelLifecycle, Sequence[ModelLifecycle]]]): Optional phase or list of phases
            in the model lifecycle to carry out the substitution for. If no `model_lifecycle` is provided, substitution
            will be carried out for all parameters.
    """
    # Search all 'value' entries in the input dict
    values_and_paths_list = search_in_config_dict(value_input, key_to_search="value")

    # Substitute the values in the config, according to their paths
    for (value, path) in values_and_paths_list:
        if metadata_keys.UI_RULES in path:
            # Skip entries that involve ui rules (these entries also have a `value` key), since we are only
            # concerned about parameter values here
            continue
        config_element = config

        for attribute_name in path:
            if attribute_name != path[-1]:
                # Traverse the config according to path
                if hasattr(config_element, attribute_name):
                    config_element = getattr(config_element, attribute_name)
                else:
                    if not allow_missing_values:
                        raise AttributeError(
                            f"Configuration does not match structure of the values to "
                            f"substitute from. The configuration {config_element} does "
                            f"not contain the parameter group {attribute_name}"
                        )
            else:
                # At the end of the path, update the attribute value
                if hasattr(config_element, attribute_name):
                    if _should_parameter_be_updated(
                        config_element=config_element,
                        parameter_name=attribute_name,
                        model_lifecycle=model_lifecycle,
                    ):
                        setattr(config_element, attribute_name, value)
                else:
                    if not allow_missing_values:
                        raise AttributeError(
                            f"Configuration does not match structure of the values "
                            f"to substitute from. Target parameter group "
                            f"{config_element} does not contain the parameter "
                            f"{attribute_name}."
                        )


def substitute_values(
    config: ConfigurableParameters,
    value_input: Union[str, DictConfig, dict, ConfigurableParameters],
    allow_missing_values: bool = False,
):
    """Substitutes values from value_input into the config object.

    The structures of value_input and config have to match in order for the values to be substituted
    correctly.

    Values are substituted in place.

    Args:
        config (ConfigurableParameters): ConfigurableParameter object to substitute values into.
        value_input (Union[str, DictConfig, dict, ConfigurableParameters]): ConfigurableParameters (either in object,
            dict, yaml or DictConfig representation) to take the values to be substituted from.
        allow_missing_values (bool): True to allow missing values in the configuration,
            i.e. if a value is found in `value_input`, but not in `config`, it will
            silently be ignored. If set to False, an AttributeError will be
            raised. Defaults to False.
    """
    # Parse input and convert to dict if needed
    if isinstance(value_input, ConfigurableParameters):
        input_dict: Dict = convert(value_input, dict)
    else:
        input_dict = input_to_config_dict(value_input, check_config_type=False)

    _substitute(config, input_dict, allow_missing_values=allow_missing_values)

    # Finally, validate the config with the updated values
    validate(config)


def substitute_values_for_lifecycle(
    config: ConfigurableParameters,
    value_input: ConfigurableParameters,
    model_lifecycle: Union[ModelLifecycle, Sequence[ModelLifecycle]],
    allow_missing_values: bool = True,
):
    """Substitutes values from value_input into the config object.

    The structures of value_input and config have to match in order for the values to be substituted correctly.
    If the argument `model_lifecycle` is provided, only parameters that affect the specified phase in the model
    lifecycle will be substituted.

    Values are substituted in place.

    Args:
        config (ConfigurableParameters): ConfigurableParameter object to substitute values into
        value_input (ConfigurableParameters): ConfigurableParameters to take the values to be substituted
            from.
        model_lifecycle (Union[ModelLifecycle, Sequence[ModelLifecycle]]): Phase or list of phases in the
            model lifecycle to carry out the substitution for. For example, if
            `model_lifecycle = ModelLifecycle.INFERENCE` is passed, only parameters that
            affect inference will be updated, and the rest of the parameters will
            remain untouched.
        allow_missing_values (bool): True to allow missing values in the configuration,
            i.e. if a value is found in `value_input`, but not in `config`, it will
            silently be ignored. If set to False, an AttributeError will be
            raised. Defaults to True.
    """
    input_dict: dict = convert(value_input, dict)

    _substitute(
        config,
        input_dict,
        allow_missing_values=allow_missing_values,
        model_lifecycle=model_lifecycle,
    )

    # Finally, validate the config with the updated values
    validate(config)



"""Unit Test for otx.algorithms.action.adapters.mmaction.data.pipeline.loading.."""

# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import pytest

from otx.algorithms.action.adapters.mmaction.data.cls_dataset import OTXActionClsDataset
from otx.algorithms.action.adapters.mmaction.data.pipelines.loading import (
    RawFrameDecode,
)
from otx.algorithms.action.configs.classification.x3d.data_pipeline import (
    train_pipeline,
)
from otx.api.entities.label import Domain
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.action.test_helpers import (
    MockPipeline,
    generate_action_cls_otx_dataset,
    generate_labels,
)


class TestRawFrameDecode:
    """Test RawFrameDecode class.

    <Steps>
        1. Create sample OTXActionClsDataset
        2. Get sample inputs from sample OTXActionClsDataset
        3. Add "frame_inds", "gt_bboxes", "proposals" attributes to sample inputs
        4. Check RawFrameDecode transform's results
            1. Whether transform creates imgs
            2. Whether transform creates proper img size
            3. Whether transform modify gt_bboxes and proposals w.r.t img_size
    """

    @pytest.fixture(autouse=True)
    def setup(self) -> None:
        self.video_len = 3
        self.frame_len = 3
        self.labels = generate_labels(3, Domain.ACTION_CLASSIFICATION)
        self.otx_dataset = generate_action_cls_otx_dataset(self.video_len, self.frame_len, self.labels)
        self.pipeline = train_pipeline
        self.dataset = OTXActionClsDataset(self.otx_dataset, self.labels, self.pipeline)
        self.dataset.pipeline = MockPipeline()

    @e2e_pytest_unit
    def test_call(self):
        """Test __call__ function."""

        inputs = self.dataset[0]
        inputs["frame_inds"] = list(range(2))
        inputs["gt_bboxes"] = np.array([[0, 0, 1, 1]])
        inputs["proposals"] = np.array([[0, 0, 1, 1]])
        decode = RawFrameDecode()
        decode.otx_dataset = self.otx_dataset
        outputs = decode(inputs)
        assert len(outputs["imgs"]) == 2
        assert outputs["original_shape"] == (256, 256)
        assert outputs["img_shape"] == (256, 256)
        assert np.all(outputs["gt_bboxes"] == np.array([[0, 0, 256, 256]]))
        assert np.all(outputs["proposals"] == np.array([[0, 0, 256, 256]]))
